import base64
import json

import boto3
import re
from freezegun import freeze_time
import sure  # noqa # pylint: disable=unused-import

from botocore.exceptions import ClientError
from unittest import SkipTest
import pytest
from moto import mock_sns, mock_sqs, settings
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from moto.core.models import responses_mock
from moto.sns import sns_backends

MESSAGE_FROM_SQS_TEMPLATE = (
    '{\n  "Message": "%s",\n  "MessageId": "%s",\n  "Signature": "EXAMPLElDMXvB8r9R83tGoNn0ecwd5UjllzsvSvbItzfaMpN2nk5HVSw7XnOn/49IkxDKz8YrlH2qJXj2iZB0Zo2O71c4qQk1fMUDi3LGpij7RCW7AW9vYYsSqIKRnFS94ilu7NFhUzLiieYr4BKHpdTmdD6c0esKEYBpabxDSc=",\n  "SignatureVersion": "1",\n  "SigningCertURL": "https://sns.us-east-1.amazonaws.com/SimpleNotificationService-f3ecfb7224c7233fe7bb5f59f96de52f.pem",\n  "Subject": "my subject",\n  "Timestamp": "2015-01-01T12:00:00.000Z",\n  "TopicArn": "arn:aws:sns:%s:'
    + ACCOUNT_ID
    + ':some-topic",\n  "Type": "Notification",\n  "UnsubscribeURL": "https://sns.us-east-1.amazonaws.com/?Action=Unsubscribe&SubscriptionArn=arn:aws:sns:us-east-1:'
    + ACCOUNT_ID
    + ':some-topic:2bcfbf39-05c3-41de-beaa-fcfcc21c8f55"\n}'
)


def to_comparable_dicts(list_entry: list):
    if list_entry:
        if isinstance(list_entry[0], dict):
            return set(map(json.dumps, list_entry))

    return set(list_entry)


@mock_sqs
@mock_sns
def test_publish_to_sqs():
    conn = boto3.client("sns", region_name="us-east-1")
    conn.create_topic(Name="some-topic")
    response = conn.list_topics()
    topic_arn = response["Topics"][0]["TopicArn"]

    sqs_conn = boto3.resource("sqs", region_name="us-east-1")
    sqs_conn.create_queue(QueueName="test-queue")

    conn.subscribe(
        TopicArn=topic_arn,
        Protocol="sqs",
        Endpoint=f"arn:aws:sqs:us-east-1:{ACCOUNT_ID}:test-queue",
    )
    message = "my message"
    with freeze_time("2015-01-01 12:00:00"):
        published_message = conn.publish(
            TopicArn=topic_arn, Message=message, Subject="my subject"
        )
    published_message_id = published_message["MessageId"]

    queue = sqs_conn.get_queue_by_name(QueueName="test-queue")
    with freeze_time("2015-01-01 12:00:01"):
        messages = queue.receive_messages(MaxNumberOfMessages=1)
    expected = MESSAGE_FROM_SQS_TEMPLATE % (message, published_message_id, "us-east-1")
    acquired_message = re.sub(
        r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}Z",
        "2015-01-01T12:00:00.000Z",
        messages[0].body,
    )
    acquired_message.should.equal(expected)


@mock_sqs
@mock_sns
def test_publish_to_sqs_raw():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(Name="some-topic")

    sqs = boto3.resource("sqs", region_name="us-east-1")
    queue = sqs.create_queue(QueueName="test-queue")

    subscription = topic.subscribe(
        Protocol="sqs", Endpoint=queue.attributes["QueueArn"]
    )

    subscription.set_attributes(
        AttributeName="RawMessageDelivery", AttributeValue="true"
    )

    message = "my message"
    with freeze_time("2015-01-01 12:00:00"):
        topic.publish(Message=message)

    with freeze_time("2015-01-01 12:00:01"):
        messages = queue.receive_messages(MaxNumberOfMessages=1)
    messages[0].body.should.equal(message)


@mock_sns
@mock_sqs
def test_publish_to_sqs_fifo():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(
        Name="topic.fifo",
        Attributes={"FifoTopic": "true", "ContentBasedDeduplication": "true"},
    )

    sqs = boto3.resource("sqs", region_name="us-east-1")
    queue = sqs.create_queue(
        QueueName="queue.fifo",
        Attributes={"FifoQueue": "true", "ContentBasedDeduplication": "true"},
    )
    topic.subscribe(Protocol="sqs", Endpoint=queue.attributes["QueueArn"])

    topic.publish(Message="message", MessageGroupId="message_group_id")


@mock_sns
@mock_sqs
def test_publish_to_sqs_fifo_with_deduplication_id():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(
        Name="topic.fifo",
        Attributes={"FifoTopic": "true"},
    )

    sqs = boto3.resource("sqs", region_name="us-east-1")
    queue = sqs.create_queue(
        QueueName="queue.fifo",
        Attributes={"FifoQueue": "true"},
    )

    topic.subscribe(
        Protocol="sqs",
        Endpoint=queue.attributes["QueueArn"],
        Attributes={"RawMessageDelivery": "true"},
    )

    message = '{"msg": "hello"}'
    with freeze_time("2015-01-01 12:00:00"):
        topic.publish(
            Message=message,
            MessageGroupId="message_group_id",
            MessageDeduplicationId="message_deduplication_id",
        )

    with freeze_time("2015-01-01 12:00:01"):
        messages = queue.receive_messages(
            MaxNumberOfMessages=1,
            AttributeNames=["MessageDeduplicationId", "MessageGroupId"],
        )
    messages[0].attributes["MessageGroupId"].should.equal("message_group_id")
    messages[0].attributes["MessageDeduplicationId"].should.equal(
        "message_deduplication_id"
    )


@mock_sns
@mock_sqs
def test_publish_to_sqs_fifo_raw_with_deduplication_id():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(
        Name="topic.fifo",
        Attributes={"FifoTopic": "true"},
    )

    sqs = boto3.resource("sqs", region_name="us-east-1")
    queue = sqs.create_queue(
        QueueName="queue.fifo",
        Attributes={"FifoQueue": "true"},
    )

    subscription = topic.subscribe(
        Protocol="sqs", Endpoint=queue.attributes["QueueArn"]
    )
    subscription.set_attributes(
        AttributeName="RawMessageDelivery", AttributeValue="true"
    )

    message = "my message"
    with freeze_time("2015-01-01 12:00:00"):
        topic.publish(
            Message=message,
            MessageGroupId="message_group_id",
            MessageDeduplicationId="message_deduplication_id",
        )

    with freeze_time("2015-01-01 12:00:01"):
        messages = queue.receive_messages(
            MaxNumberOfMessages=1,
            AttributeNames=["MessageDeduplicationId", "MessageGroupId"],
        )
    messages[0].attributes["MessageGroupId"].should.equal("message_group_id")
    messages[0].attributes["MessageDeduplicationId"].should.equal(
        "message_deduplication_id"
    )


@mock_sqs
@mock_sns
def test_publish_to_sqs_bad():
    conn = boto3.client("sns", region_name="us-east-1")
    conn.create_topic(Name="some-topic")
    response = conn.list_topics()
    topic_arn = response["Topics"][0]["TopicArn"]

    sqs_conn = boto3.resource("sqs", region_name="us-east-1")
    sqs_conn.create_queue(QueueName="test-queue")

    conn.subscribe(
        TopicArn=topic_arn,
        Protocol="sqs",
        Endpoint=f"arn:aws:sqs:us-east-1:{ACCOUNT_ID}:test-queue",
    )
    message = "my message"
    try:
        # Test missing Value
        conn.publish(
            TopicArn=topic_arn,
            Message=message,
            MessageAttributes={"store": {"DataType": "String"}},
        )
    except ClientError as err:
        err.response["Error"]["Code"].should.equal("InvalidParameterValue")
    try:
        # Test empty DataType (if the DataType field is missing entirely
        # botocore throws an exception during validation)
        conn.publish(
            TopicArn=topic_arn,
            Message=message,
            MessageAttributes={
                "store": {"DataType": "", "StringValue": "example_corp"}
            },
        )
    except ClientError as err:
        err.response["Error"]["Code"].should.equal("InvalidParameterValue")
    try:
        # Test empty Value
        conn.publish(
            TopicArn=topic_arn,
            Message=message,
            MessageAttributes={"store": {"DataType": "String", "StringValue": ""}},
        )
    except ClientError as err:
        err.response["Error"]["Code"].should.equal("InvalidParameterValue")
    try:
        # Test Number DataType, with a non numeric value
        conn.publish(
            TopicArn=topic_arn,
            Message=message,
            MessageAttributes={"price": {"DataType": "Number", "StringValue": "error"}},
        )
    except ClientError as err:
        err.response["Error"]["Code"].should.equal("InvalidParameterValue")
        err.response["Error"]["Message"].should.equal(
            "An error occurred (ParameterValueInvalid) when calling the Publish operation: Could not cast message attribute 'price' value to number."
        )


@mock_sqs
@mock_sns
def test_publish_to_sqs_msg_attr_byte_value():
    conn = boto3.client("sns", region_name="us-east-1")
    conn.create_topic(Name="some-topic")
    response = conn.list_topics()
    topic_arn = response["Topics"][0]["TopicArn"]
    sqs = boto3.resource("sqs", region_name="us-east-1")
    queue = sqs.create_queue(QueueName="test-queue")
    conn.subscribe(
        TopicArn=topic_arn, Protocol="sqs", Endpoint=queue.attributes["QueueArn"]
    )
    queue_raw = sqs.create_queue(QueueName="test-queue-raw")
    conn.subscribe(
        TopicArn=topic_arn,
        Protocol="sqs",
        Endpoint=queue_raw.attributes["QueueArn"],
        Attributes={"RawMessageDelivery": "true"},
    )

    conn.publish(
        TopicArn=topic_arn,
        Message="my message",
        MessageAttributes={
            "store": {"DataType": "Binary", "BinaryValue": b"\x02\x03\x04"}
        },
    )

    message = json.loads(queue.receive_messages()[0].body)
    message["Message"].should.equal("my message")
    message["MessageAttributes"].should.equal(
        {
            "store": {
                "Type": "Binary",
                "Value": base64.b64encode(b"\x02\x03\x04").decode(),
            }
        }
    )

    message = queue_raw.receive_messages()[0]
    message.body.should.equal("my message")


@mock_sqs
@mock_sns
def test_publish_to_sqs_msg_attr_number_type():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(Name="test-topic")
    sqs = boto3.resource("sqs", region_name="us-east-1")
    queue = sqs.create_queue(QueueName="test-queue")
    topic.subscribe(Protocol="sqs", Endpoint=queue.attributes["QueueArn"])
    queue_raw = sqs.create_queue(QueueName="test-queue-raw")
    topic.subscribe(
        Protocol="sqs",
        Endpoint=queue_raw.attributes["QueueArn"],
        Attributes={"RawMessageDelivery": "true"},
    )

    topic.publish(
        Message="test message",
        MessageAttributes={"retries": {"DataType": "Number", "StringValue": "0"}},
    )

    message = json.loads(queue.receive_messages()[0].body)
    message["Message"].should.equal("test message")
    message["MessageAttributes"].should.equal(
        {"retries": {"Type": "Number", "Value": "0"}}
    )

    message = queue_raw.receive_messages()[0]
    message.body.should.equal("test message")


@mock_sqs
@mock_sns
def test_publish_to_sqs_msg_attr_different_formats():
    """
    Verify different Number-formats are processed correctly
    """
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(Name="test-topic")
    sqs = boto3.resource("sqs", region_name="us-east-1")
    sqs_client = boto3.client("sqs", region_name="us-east-1")
    queue_raw = sqs.create_queue(QueueName="test-queue-raw")

    topic.subscribe(
        Protocol="sqs",
        Endpoint=queue_raw.attributes["QueueArn"],
        Attributes={"RawMessageDelivery": "true"},
    )

    topic.publish(
        Message="test message",
        MessageAttributes={
            "integer": {"DataType": "Number", "StringValue": "123"},
            "float": {"DataType": "Number", "StringValue": "12.34"},
            "big-integer": {"DataType": "Number", "StringValue": "123456789"},
            "big-float": {"DataType": "Number", "StringValue": "123456.789"},
        },
    )

    messages_resp = sqs_client.receive_message(
        QueueUrl=queue_raw.url, MessageAttributeNames=["All"]
    )
    message = messages_resp["Messages"][0]
    message_attributes = message["MessageAttributes"]
    message_attributes.should.equal(
        {
            "integer": {"DataType": "Number", "StringValue": "123"},
            "float": {"DataType": "Number", "StringValue": "12.34"},
            "big-integer": {"DataType": "Number", "StringValue": "123456789"},
            "big-float": {"DataType": "Number", "StringValue": "123456.789"},
        }
    )


@mock_sns
def test_publish_sms():
    client = boto3.client("sns", region_name="us-east-1")

    result = client.publish(PhoneNumber="+15551234567", Message="my message")

    result.should.contain("MessageId")
    if not settings.TEST_SERVER_MODE:
        sns_backend = sns_backends[ACCOUNT_ID]["us-east-1"]
        sns_backend.sms_messages.should.have.key(result["MessageId"]).being.equal(
            ("+15551234567", "my message")
        )


@mock_sns
def test_publish_bad_sms():
    client = boto3.client("sns", region_name="us-east-1")

    # Test invalid number
    with pytest.raises(ClientError) as cm:
        client.publish(PhoneNumber="NAA+15551234567", Message="my message")
    cm.value.response["Error"]["Code"].should.equal("InvalidParameter")
    cm.value.response["Error"]["Message"].should.contain("not meet the E164")

    # Test to long ASCII message
    with pytest.raises(ClientError) as cm:
        client.publish(PhoneNumber="+15551234567", Message="a" * 1601)
    cm.value.response["Error"]["Code"].should.equal("InvalidParameter")
    cm.value.response["Error"]["Message"].should.contain("must be less than 1600")


@mock_sqs
@mock_sns
def test_publish_to_sqs_dump_json():
    conn = boto3.client("sns", region_name="us-east-1")
    conn.create_topic(Name="some-topic")
    response = conn.list_topics()
    topic_arn = response["Topics"][0]["TopicArn"]

    sqs_conn = boto3.resource("sqs", region_name="us-east-1")
    sqs_conn.create_queue(QueueName="test-queue")

    conn.subscribe(
        TopicArn=topic_arn,
        Protocol="sqs",
        Endpoint=f"arn:aws:sqs:us-east-1:{ACCOUNT_ID}:test-queue",
    )

    message = json.dumps(
        {
            "Records": [
                {
                    "eventVersion": "2.0",
                    "eventSource": "aws:s3",
                    "s3": {"s3SchemaVersion": "1.0"},
                }
            ]
        },
        sort_keys=True,
    )
    with freeze_time("2015-01-01 12:00:00"):
        published_message = conn.publish(
            TopicArn=topic_arn, Message=message, Subject="my subject"
        )
    published_message_id = published_message["MessageId"]

    queue = sqs_conn.get_queue_by_name(QueueName="test-queue")
    with freeze_time("2015-01-01 12:00:01"):
        messages = queue.receive_messages(MaxNumberOfMessages=1)

    escaped = message.replace('"', '\\"')
    expected = MESSAGE_FROM_SQS_TEMPLATE % (escaped, published_message_id, "us-east-1")
    acquired_message = re.sub(
        r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}Z",
        "2015-01-01T12:00:00.000Z",
        messages[0].body,
    )
    acquired_message.should.equal(expected)


@mock_sqs
@mock_sns
def test_publish_to_sqs_in_different_region():
    conn = boto3.client("sns", region_name="us-west-1")
    conn.create_topic(Name="some-topic")
    response = conn.list_topics()
    topic_arn = response["Topics"][0]["TopicArn"]

    sqs_conn = boto3.resource("sqs", region_name="us-west-2")
    sqs_conn.create_queue(QueueName="test-queue")

    conn.subscribe(
        TopicArn=topic_arn,
        Protocol="sqs",
        Endpoint=f"arn:aws:sqs:us-west-2:{ACCOUNT_ID}:test-queue",
    )

    message = "my message"
    with freeze_time("2015-01-01 12:00:00"):
        published_message = conn.publish(
            TopicArn=topic_arn, Message=message, Subject="my subject"
        )
    published_message_id = published_message["MessageId"]

    queue = sqs_conn.get_queue_by_name(QueueName="test-queue")
    with freeze_time("2015-01-01 12:00:01"):
        messages = queue.receive_messages(MaxNumberOfMessages=1)
    expected = MESSAGE_FROM_SQS_TEMPLATE % (message, published_message_id, "us-west-1")
    acquired_message = re.sub(
        r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{3}Z",
        "2015-01-01T12:00:00.000Z",
        messages[0].body,
    )
    acquired_message.should.equal(expected)


@freeze_time("2013-01-01")
@mock_sns
def test_publish_to_http():
    if settings.TEST_SERVER_MODE:
        raise SkipTest("Can't mock requests in ServerMode")

    def callback(request):
        request.headers["Content-Type"].should.equal("text/plain; charset=UTF-8")
        json.loads.when.called_with(request.body.decode()).should_not.throw(Exception)
        return 200, {}, ""

    responses_mock.add_callback(
        method="POST", url="http://example.com/foobar", callback=callback
    )

    conn = boto3.client("sns", region_name="us-east-1")
    conn.create_topic(Name="some-topic")
    response = conn.list_topics()
    topic_arn = response["Topics"][0]["TopicArn"]

    conn.subscribe(
        TopicArn=topic_arn, Protocol="http", Endpoint="http://example.com/foobar"
    )

    conn.publish(TopicArn=topic_arn, Message="my message", Subject="my subject")

    sns_backend = sns_backends[ACCOUNT_ID]["us-east-1"]
    sns_backend.topics[topic_arn].sent_notifications.should.have.length_of(1)
    notification = sns_backend.topics[topic_arn].sent_notifications[0]
    _, msg, subject, _, _ = notification
    msg.should.equal("my message")
    subject.should.equal("my subject")


@mock_sqs
@mock_sns
def test_publish_subject():
    conn = boto3.client("sns", region_name="us-east-1")
    conn.create_topic(Name="some-topic")
    response = conn.list_topics()
    topic_arn = response["Topics"][0]["TopicArn"]

    sqs_conn = boto3.resource("sqs", region_name="us-east-1")
    sqs_conn.create_queue(QueueName="test-queue")

    conn.subscribe(
        TopicArn=topic_arn,
        Protocol="sqs",
        Endpoint=f"arn:aws:sqs:us-east-1:{ACCOUNT_ID}:test-queue",
    )
    message = "my message"
    subject1 = "test subject"
    subject2 = "test subject" * 20
    with freeze_time("2015-01-01 12:00:00"):
        conn.publish(TopicArn=topic_arn, Message=message, Subject=subject1)

    # Just that it doesnt error is a pass
    try:
        with freeze_time("2015-01-01 12:00:00"):
            conn.publish(TopicArn=topic_arn, Message=message, Subject=subject2)
    except ClientError as err:
        err.response["Error"]["Code"].should.equal("InvalidParameter")
    else:
        raise RuntimeError("Should have raised an InvalidParameter exception")


@mock_sqs
@mock_sns
def test_publish_null_subject():
    conn = boto3.client("sns", region_name="us-east-1")
    conn.create_topic(Name="some-topic")
    response = conn.list_topics()
    topic_arn = response["Topics"][0]["TopicArn"]

    sqs_conn = boto3.resource("sqs", region_name="us-east-1")
    sqs_conn.create_queue(QueueName="test-queue")

    conn.subscribe(
        TopicArn=topic_arn,
        Protocol="sqs",
        Endpoint=f"arn:aws:sqs:us-east-1:{ACCOUNT_ID}:test-queue",
    )
    message = "my message"
    with freeze_time("2015-01-01 12:00:00"):
        conn.publish(TopicArn=topic_arn, Message=message)

    queue = sqs_conn.get_queue_by_name(QueueName="test-queue")
    with freeze_time("2015-01-01 12:00:01"):
        messages = queue.receive_messages(MaxNumberOfMessages=1)

    acquired_message = json.loads(messages[0].body)
    acquired_message["Message"].should.equal(message)
    acquired_message.shouldnt.have.key("Subject")


@mock_sns
def test_publish_message_too_long():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(Name="some-topic")

    with pytest.raises(ClientError):
        topic.publish(Message="".join(["." for i in range(0, 262145)]))

    # message short enough - does not raise an error
    topic.publish(Message="".join(["." for i in range(0, 262144)]))


@mock_sns
def test_publish_fifo_needs_group_id():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(
        Name="topic.fifo",
        Attributes={"FifoTopic": "true", "ContentBasedDeduplication": "true"},
    )

    with pytest.raises(
        ClientError, match="The request must contain the parameter MessageGroupId"
    ):
        topic.publish(Message="message")

    # message group included - OK
    topic.publish(Message="message", MessageGroupId="message_group_id")


@mock_sns
@mock_sqs
def test_publish_group_id_to_non_fifo():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(Name="topic")

    with pytest.raises(
        ClientError,
        match="The request includes MessageGroupId parameter that is not valid for this topic type",
    ):
        topic.publish(Message="message", MessageGroupId="message_group_id")

    # message group not included - OK
    topic.publish(Message="message")


@mock_sns
def test_publish_fifo_needs_deduplication_id():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(
        Name="topic.fifo",
        Attributes={"FifoTopic": "true"},
    )

    with pytest.raises(
        ClientError,
        match="The topic should either have ContentBasedDeduplication enabled or MessageDeduplicationId provided explicitly",
    ):
        topic.publish(Message="message", MessageGroupId="message_group_id")

    # message deduplication id included - OK
    topic.publish(
        Message="message",
        MessageGroupId="message_group_id",
        MessageDeduplicationId="message_deduplication_id",
    )


@mock_sns
def test_publish_deduplication_id_to_non_fifo():
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(Name="topic")

    with pytest.raises(
        ClientError,
        match="The request includes MessageDeduplicationId parameter that is not valid for this topic type",
    ):
        topic.publish(
            Message="message", MessageDeduplicationId="message_deduplication_id"
        )

    # message group not included - OK
    topic.publish(Message="message")


def _setup_filter_policy_test(
    filter_policy: dict, filter_policy_scope: str = "MessageAttributes"
):
    sns = boto3.resource("sns", region_name="us-east-1")
    topic = sns.create_topic(Name="some-topic")

    sqs = boto3.resource("sqs", region_name="us-east-1")
    queue = sqs.create_queue(QueueName="test-queue")

    subscription = topic.subscribe(
        Protocol="sqs", Endpoint=queue.attributes["QueueArn"]
    )

    subscription.set_attributes(
        AttributeName="FilterPolicyScope", AttributeValue=filter_policy_scope
    )

    subscription.set_attributes(
        AttributeName="FilterPolicy", AttributeValue=json.dumps(filter_policy)
    )

    return topic, queue


@mock_sqs
@mock_sns
def test_filtering_exact_string():
    topic, queue = _setup_filter_policy_test({"store": ["example_corp"]})

    topic.publish(
        Message="match",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [{"store": {"Type": "String", "Value": "example_corp"}}]
    )


@mock_sqs
@mock_sns
def test_filtering_exact_string_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"store": "example_corp"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"result": {"Type": "String", "Value": "match"}}])
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"store": "example_corp"}])


@mock_sqs
@mock_sns
def test_filtering_exact_string_multiple_message_attributes():
    topic, queue = _setup_filter_policy_test({"store": ["example_corp"]})

    topic.publish(
        Message="match",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"},
            "event": {"DataType": "String", "StringValue": "order_cancelled"},
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "store": {"Type": "String", "Value": "example_corp"},
                "event": {"Type": "String", "Value": "order_cancelled"},
            }
        ]
    )


@mock_sqs
@mock_sns
def test_filtering_exact_string_multiple_message_attributes_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"store": "example_corp", "event": "order_cancelled"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"result": {"Type": "String", "Value": "match"}}])
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"store": "example_corp", "event": "order_cancelled"}])


@mock_sqs
@mock_sns
def test_filtering_exact_string_OR_matching():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp", "different_corp"]}
    )

    topic.publish(
        Message="match example_corp",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"}
        },
    )
    topic.publish(
        Message="match different_corp",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "different_corp"}
        },
    )
    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match example_corp", "match different_corp"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {"store": {"Type": "String", "Value": "example_corp"}},
            {"store": {"Type": "String", "Value": "different_corp"}},
        ]
    )


@mock_sqs
@mock_sns
def test_filtering_exact_string_OR_matching_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp", "different_corp"]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"store": "example_corp"}),
        MessageAttributes={
            "result": {"DataType": "String", "StringValue": "match example_corp"}
        },
    )

    topic.publish(
        Message=json.dumps({"store": "different_corp"}),
        MessageAttributes={
            "result": {"DataType": "String", "StringValue": "match different_corp"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    to_comparable_dicts(message_attributes).should.equal(
        to_comparable_dicts(
            [
                {"result": {"Type": "String", "Value": "match example_corp"}},
                {"result": {"Type": "String", "Value": "match different_corp"}},
            ]
        )
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"store": "example_corp"}, {"store": "different_corp"}])
    )


@mock_sqs
@mock_sns
def test_filtering_exact_string_AND_matching_positive():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"], "event": ["order_cancelled"]}
    )

    topic.publish(
        Message="match example_corp order_cancelled",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"},
            "event": {"DataType": "String", "StringValue": "order_cancelled"},
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match example_corp order_cancelled"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "store": {"Type": "String", "Value": "example_corp"},
                "event": {"Type": "String", "Value": "order_cancelled"},
            }
        ]
    )


@mock_sqs
@mock_sns
def test_filtering_exact_string_AND_matching_positive_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"], "event": ["order_cancelled"]},
        filter_policy_scope="MessageBody",
    )

    topic.publish(
        Message=json.dumps({"store": "example_corp", "event": "order_cancelled"}),
        MessageAttributes={
            "result": {
                "DataType": "String",
                "StringValue": "match example_corp order_cancelled",
            }
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {
                    "Type": "String",
                    "Value": "match example_corp order_cancelled",
                },
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"store": "example_corp", "event": "order_cancelled"}])


@mock_sqs
@mock_sns
def test_filtering_exact_string_AND_matching_no_match():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"], "event": ["order_cancelled"]}
    )

    topic.publish(
        Message="match example_corp order_accepted",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"},
            "event": {"DataType": "String", "StringValue": "order_accepted"},
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_string_AND_matching_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"], "event": ["order_cancelled"]},
        filter_policy_scope="MessageBody",
    )

    topic.publish(
        Message=json.dumps({"store": "example_corp", "event": "order_accepted"}),
        MessageAttributes={
            "result": {
                "DataType": "String",
                "StringValue": "match example_corp order_accepted",
            }
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_string_no_match():
    topic, queue = _setup_filter_policy_test({"store": ["example_corp"]})

    topic.publish(
        Message="no match",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "different_corp"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_string_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"store": "different_corp"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_string_no_attributes_no_match():
    topic, queue = _setup_filter_policy_test({"store": ["example_corp"]})

    topic.publish(Message="no match")

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_string_empty_body_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_number_int():
    topic, queue = _setup_filter_policy_test({"price": [100]})

    topic.publish(
        Message="match",
        MessageAttributes={"price": {"DataType": "Number", "StringValue": "100"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"price": {"Type": "Number", "Value": "100"}}])


@mock_sqs
@mock_sns
def test_filtering_exact_number_int_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": 100}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"price": 100}])


@mock_sqs
@mock_sns
def test_filtering_exact_number_float():
    topic, queue = _setup_filter_policy_test({"price": [100.1]})

    topic.publish(
        Message="match",
        MessageAttributes={"price": {"DataType": "Number", "StringValue": "100.1"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"price": {"Type": "Number", "Value": "100.1"}}])


@mock_sqs
@mock_sns
def test_filtering_exact_number_float_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100.1]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": 100.1}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"price": 100.1}])


@mock_sqs
@mock_sns
def test_filtering_exact_number_float_accuracy():
    topic, queue = _setup_filter_policy_test({"price": [100.123456789]})

    topic.publish(
        Message="match",
        MessageAttributes={
            "price": {"DataType": "Number", "StringValue": "100.1234567"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [{"price": {"Type": "Number", "Value": "100.1234567"}}]
    )


@mock_sqs
@mock_sns
def test_filtering_exact_number_float_accuracy_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100.123456789]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": 100.1234567}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"price": 100.1234567}])


@mock_sqs
@mock_sns
def test_filtering_exact_number_no_match():
    topic, queue = _setup_filter_policy_test({"price": [100]})

    topic.publish(
        Message="no match",
        MessageAttributes={"price": {"DataType": "Number", "StringValue": "101"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_number_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": 101}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_number_with_string_no_match():
    topic, queue = _setup_filter_policy_test({"price": [100]})

    topic.publish(
        Message="no match",
        MessageAttributes={"price": {"DataType": "String", "StringValue": "100"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_number_with_string_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": "100"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_string_array_match():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": ["basketball", "baseball"]}
    )

    topic.publish(
        Message="match",
        MessageAttributes={
            "customer_interests": {
                "DataType": "String.Array",
                "StringValue": json.dumps(["basketball", "rugby"]),
            }
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "customer_interests": {
                    "Type": "String.Array",
                    "Value": json.dumps(["basketball", "rugby"]),
                }
            }
        ]
    )


@mock_sqs
@mock_sns
def test_filtering_string_array_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": ["basketball", "baseball"]},
        filter_policy_scope="MessageBody",
    )

    topic.publish(
        Message=json.dumps({"customer_interests": ["basketball", "rugby"]}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"customer_interests": ["basketball", "rugby"]}])


@mock_sqs
@mock_sns
def test_filtering_string_array_no_match():
    topic, queue = _setup_filter_policy_test({"customer_interests": ["baseball"]})

    topic.publish(
        Message="no_match",
        MessageAttributes={
            "customer_interests": {
                "DataType": "String.Array",
                "StringValue": json.dumps(["basketball", "rugby"]),
            }
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_string_array_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": ["baseball"]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"customer_interests": ["basketball", "rugby"]}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no_match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_string_array_with_number_match():
    topic, queue = _setup_filter_policy_test({"price": [100, 500]})

    topic.publish(
        Message="match",
        MessageAttributes={
            "price": {"DataType": "String.Array", "StringValue": json.dumps([100, 50])}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [{"price": {"Type": "String.Array", "Value": json.dumps([100, 50])}}]
    )


@mock_sqs
@mock_sns
def test_filtering_string_array_with_number_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100, 500]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": [100, 50]}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"price": [100, 50]}])


@mock_sqs
@mock_sns
def test_filtering_string_array_with_number_float_accuracy_match():
    topic, queue = _setup_filter_policy_test({"price": [100.123456789, 500]})

    topic.publish(
        Message="match",
        MessageAttributes={
            "price": {
                "DataType": "String.Array",
                "StringValue": json.dumps([100.1234567, 50]),
            }
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [{"price": {"Type": "String.Array", "Value": json.dumps([100.1234567, 50])}}]
    )


@mock_sqs
@mock_sns
def test_filtering_string_array_with_number_float_accuracy_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100.123456789, 500]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": [100.1234567, 50]}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"price": [100.1234567, 50]}])


@mock_sqs
@mock_sns
# this is the correct behavior from SNS
def test_filtering_string_array_with_number_no_array_match():
    topic, queue = _setup_filter_policy_test({"price": [100, 500]})

    topic.publish(
        Message="match",
        MessageAttributes={"price": {"DataType": "String.Array", "StringValue": "100"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [{"price": {"Type": "String.Array", "Value": "100"}}]
    )


@mock_sqs
@mock_sns
# this is the correct behavior from SNS
def test_filtering_string_array_with_number_no_array_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100, 500]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": 100}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"price": 100}])


@mock_sqs
@mock_sns
def test_filtering_string_array_with_number_no_match():
    topic, queue = _setup_filter_policy_test({"price": [500]})

    topic.publish(
        Message="no_match",
        MessageAttributes={
            "price": {"DataType": "String.Array", "StringValue": json.dumps([100, 50])}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_string_array_with_number_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [500]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": [100, 50]}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no_match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
# this is the correct behavior from SNS
def test_filtering_string_array_with_string_no_array_no_match():
    topic, queue = _setup_filter_policy_test({"price": [100]})

    topic.publish(
        Message="no_match",
        MessageAttributes={
            "price": {"DataType": "String.Array", "StringValue": "one hundred"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_string_array_with_string_no_array_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"price": [100]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"price": "one hundred"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no_match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_attribute_key_exists_match():
    topic, queue = _setup_filter_policy_test({"store": [{"exists": True}]})

    topic.publish(
        Message="match",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [{"store": {"Type": "String", "Value": "example_corp"}}]
    )


@mock_sqs
@mock_sns
def test_filtering_body_key_exists_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": [{"exists": True}]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"store": "example_corp"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"store": "example_corp"}])


@mock_sqs
@mock_sns
def test_filtering_attribute_key_exists_no_match():
    topic, queue = _setup_filter_policy_test({"store": [{"exists": True}]})

    topic.publish(
        Message="no match",
        MessageAttributes={
            "event": {"DataType": "String", "StringValue": "order_cancelled"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_body_key_exists_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": [{"exists": True}]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"event": "order_cancelled"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_attribute_key_not_exists_match():
    topic, queue = _setup_filter_policy_test({"store": [{"exists": False}]})

    topic.publish(
        Message="match",
        MessageAttributes={
            "event": {"DataType": "String", "StringValue": "order_cancelled"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [{"event": {"Type": "String", "Value": "order_cancelled"}}]
    )


@mock_sqs
@mock_sns
def test_filtering_body_key_not_exists_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": [{"exists": False}]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"event": "order_cancelled"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([{"event": "order_cancelled"}])


@mock_sqs
@mock_sns
def test_filtering_attribute_key_not_exists_no_match():
    topic, queue = _setup_filter_policy_test({"store": [{"exists": False}]})

    topic.publish(
        Message="no match",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"}
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_body_key_not_exists_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {"store": [{"exists": False}]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"store": "example_corp"}),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_all_AND_matching_match():
    topic, queue = _setup_filter_policy_test(
        {
            "store": [{"exists": True}],
            "event": ["order_cancelled"],
            "customer_interests": ["basketball", "baseball"],
            "price": [100],
        }
    )

    topic.publish(
        Message="match",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"},
            "event": {"DataType": "String", "StringValue": "order_cancelled"},
            "customer_interests": {
                "DataType": "String.Array",
                "StringValue": json.dumps(["basketball", "rugby"]),
            },
            "price": {"DataType": "Number", "StringValue": "100"},
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal(["match"])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "store": {"Type": "String", "Value": "example_corp"},
                "event": {"Type": "String", "Value": "order_cancelled"},
                "customer_interests": {
                    "Type": "String.Array",
                    "Value": json.dumps(["basketball", "rugby"]),
                },
                "price": {"Type": "Number", "Value": "100"},
            }
        ]
    )


@mock_sqs
@mock_sns
def test_filtering_all_AND_matching_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "store": [{"exists": True}],
            "event": ["order_cancelled"],
            "customer_interests": ["basketball", "baseball"],
            "price": [100],
        },
        filter_policy_scope="MessageBody",
    )

    topic.publish(
        Message=json.dumps(
            {
                "store": "example_corp",
                "event": "order_cancelled",
                "customer_interests": ["basketball", "rugby"],
                "price": 100,
            }
        ),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal(
        [
            {
                "result": {"Type": "String", "Value": "match"},
            }
        ]
    )
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal(
        [
            {
                "store": "example_corp",
                "event": "order_cancelled",
                "customer_interests": ["basketball", "rugby"],
                "price": 100,
            }
        ]
    )


@mock_sqs
@mock_sns
def test_filtering_all_AND_matching_no_match():
    topic, queue = _setup_filter_policy_test(
        {
            "store": [{"exists": True}],
            "event": ["order_cancelled"],
            "customer_interests": ["basketball", "baseball"],
            "price": [100],
            "encrypted": [False],
        }
    )

    topic.publish(
        Message="no match",
        MessageAttributes={
            "store": {"DataType": "String", "StringValue": "example_corp"},
            "event": {"DataType": "String", "StringValue": "order_cancelled"},
            "customer_interests": {
                "DataType": "String.Array",
                "StringValue": json.dumps(["basketball", "rugby"]),
            },
            "price": {"DataType": "Number", "StringValue": "100"},
        },
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_all_AND_matching_no_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "store": [{"exists": True}],
            "event": ["order_cancelled"],
            "customer_interests": ["basketball", "baseball"],
            "price": [100],
            "encrypted": [False],
        },
        filter_policy_scope="MessageBody",
    )

    topic.publish(
        Message=json.dumps(
            {
                "store": "example_corp",
                "event": "order_cancelled",
                "customer_interests": ["basketball", "rugby"],
                "price": 100,
            }
        ),
        MessageAttributes={"result": {"DataType": "String", "StringValue": "no match"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_prefix():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": [{"prefix": "bas"}]}
    )

    for interest, idx in [("basketball", "1"), ("rugby", "2"), ("baseball", "3")]:
        topic.publish(
            Message=f"match{idx}",
            MessageAttributes={
                "customer_interests": {"DataType": "String", "StringValue": interest},
            },
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    set(message_bodies).should.equal({"match1", "match3"})


@mock_sqs
@mock_sns
def test_filtering_prefix_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "customer_interests": [{"prefix": "bas"}],
        },
        filter_policy_scope="MessageBody",
    )

    for interest in ["basketball", "rugby", "baseball"]:
        topic.publish(
            Message=json.dumps(
                {
                    "customer_interests": interest,
                }
            )
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts(
            [{"customer_interests": "basketball"}, {"customer_interests": "baseball"}]
        )
    )


@mock_sqs
@mock_sns
def test_filtering_anything_but():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": [{"anything-but": "basketball"}]}
    )

    for interest, idx in [("basketball", "1"), ("rugby", "2"), ("baseball", "3")]:
        topic.publish(
            Message=f"match{idx}",
            MessageAttributes={
                "customer_interests": {"DataType": "String", "StringValue": interest},
            },
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    set(message_bodies).should.equal({"match2", "match3"})


@mock_sqs
@mock_sns
def test_filtering_anything_but_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "customer_interests": [{"anything-but": "basketball"}],
        },
        filter_policy_scope="MessageBody",
    )

    for interest in ["basketball", "rugby", "baseball"]:
        topic.publish(
            Message=json.dumps(
                {
                    "customer_interests": interest,
                }
            )
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts(
            [{"customer_interests": "rugby"}, {"customer_interests": "baseball"}]
        )
    )


@mock_sqs
@mock_sns
def test_filtering_anything_but_multiple_values():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": [{"anything-but": ["basketball", "rugby"]}]}
    )

    for interest, idx in [("basketball", "1"), ("rugby", "2"), ("baseball", "3")]:
        topic.publish(
            Message=f"match{idx}",
            MessageAttributes={
                "customer_interests": {"DataType": "String", "StringValue": interest},
            },
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    set(message_bodies).should.equal({"match3"})


@mock_sqs
@mock_sns
def test_filtering_anything_but_multiple_values_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "customer_interests": [{"anything-but": ["basketball", "rugby"]}],
        },
        filter_policy_scope="MessageBody",
    )

    for interest in ["basketball", "rugby", "baseball"]:
        topic.publish(
            Message=json.dumps(
                {
                    "customer_interests": interest,
                }
            )
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"customer_interests": "baseball"}])
    )


@mock_sqs
@mock_sns
def test_filtering_anything_but_prefix():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": [{"anything-but": {"prefix": "bas"}}]}
    )

    for interest, idx in [("basketball", "1"), ("rugby", "2"), ("baseball", "3")]:
        topic.publish(
            Message=f"match{idx}",
            MessageAttributes={
                "customer_interests": {"DataType": "String", "StringValue": interest},
            },
        )

    # This should match rugby only
    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    set(message_bodies).should.equal({"match2"})


@mock_sqs
@mock_sns
def test_filtering_anything_but_prefix_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "customer_interests": [{"anything-but": {"prefix": "bas"}}],
        },
        filter_policy_scope="MessageBody",
    )

    for interest in ["basketball", "rugby", "baseball"]:
        topic.publish(
            Message=json.dumps(
                {
                    "customer_interests": interest,
                }
            )
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"customer_interests": "rugby"}])
    )


@mock_sqs
@mock_sns
def test_filtering_anything_but_unknown():
    try:
        _setup_filter_policy_test(
            {"customer_interests": [{"anything-but": {"unknown": "bas"}}]}
        )
    except ClientError as err:
        err.response["Error"]["Code"].should.equal("InvalidParameter")


@mock_sqs
@mock_sns
def test_filtering_anything_but_unknown_message_body_raises():
    try:
        _setup_filter_policy_test(
            {
                "customer_interests": [{"anything-but": {"unknown": "bas"}}],
            },
            filter_policy_scope="MessageBody",
        )
    except ClientError as err:
        err.response["Error"]["Code"].should.equal("InvalidParameter")


@mock_sqs
@mock_sns
def test_filtering_anything_but_numeric():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": [{"anything-but": [100]}]}
    )

    for nr, idx in [("50", "1"), ("100", "2"), ("150", "3")]:
        topic.publish(
            Message=f"match{idx}",
            MessageAttributes={
                "customer_interests": {"DataType": "Number", "StringValue": nr},
            },
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    set(message_bodies).should.equal({"match1", "match3"})


@mock_sqs
@mock_sns
def test_filtering_anything_but_numeric_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "customer_interests": [{"anything-but": [100]}],
        },
        filter_policy_scope="MessageBody",
    )

    for nr in [50, 100, 150]:
        topic.publish(
            Message=json.dumps(
                {
                    "customer_interests": nr,
                }
            )
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)

    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"customer_interests": 50}, {"customer_interests": 150}])
    )


@mock_sqs
@mock_sns
def test_filtering_anything_but_numeric_string():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": [{"anything-but": ["100"]}]}
    )

    for nr, idx in [("50", "1"), ("100", "2"), ("150", "3")]:
        topic.publish(
            Message=f"match{idx}",
            MessageAttributes={
                "customer_interests": {"DataType": "Number", "StringValue": nr},
            },
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    set(message_bodies).should.equal({"match1", "match2", "match3"})


@mock_sqs
@mock_sns
def test_filtering_anything_but_numeric_string_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "customer_interests": [{"anything-but": ["100"]}],
        },
        filter_policy_scope="MessageBody",
    )

    for nr in [50, 100, 150]:
        topic.publish(
            Message=json.dumps(
                {
                    "customer_interests": nr,
                }
            )
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)

    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts(
            [
                {"customer_interests": 50},
                {"customer_interests": 100},
                {"customer_interests": 150},
            ]
        )
    )


@mock_sqs
@mock_sns
def test_filtering_numeric_match():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": [{"numeric": ["=", 100]}]}
    )

    for nr, idx in [("50", "1"), ("100", "2"), ("150", "3")]:
        topic.publish(
            Message=f"match{idx}",
            MessageAttributes={
                "customer_interests": {"DataType": "Number", "StringValue": nr},
            },
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    set(message_bodies).should.equal({"match2"})


@mock_sqs
@mock_sns
def test_filtering_numeric_match_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "customer_interests": [{"numeric": ["=", 100]}],
        },
        filter_policy_scope="MessageBody",
    )

    for nr in [50, 100, 150]:
        topic.publish(
            Message=json.dumps(
                {
                    "customer_interests": nr,
                }
            )
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"customer_interests": 100}])
    )


@mock_sqs
@mock_sns
def test_filtering_numeric_range():
    topic, queue = _setup_filter_policy_test(
        {"customer_interests": [{"numeric": [">", 49, "<=", 100]}]}
    )

    for nr, idx in [("50", "1"), ("100", "2"), ("150", "3")]:
        topic.publish(
            Message=f"match{idx}",
            MessageAttributes={
                "customer_interests": {"DataType": "Number", "StringValue": nr},
            },
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    set(message_bodies).should.equal({"match1", "match2"})


@mock_sqs
@mock_sns
def test_filtering_numeric_range_message_body():
    topic, queue = _setup_filter_policy_test(
        {
            "customer_interests": [{"numeric": [">", 49, "<=", 100]}],
        },
        filter_policy_scope="MessageBody",
    )

    for nr in [50, 100, 150]:
        topic.publish(
            Message=json.dumps(
                {
                    "customer_interests": nr,
                }
            )
        )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"customer_interests": 50}, {"customer_interests": 100}])
    )


@mock_sqs
@mock_sns
def test_filtering_exact_string_message_body_invalid_json_no_match():
    topic, queue = _setup_filter_policy_test(
        {"store": ["example_corp"]}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message='{"store": "another_corp"',
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_exact_string_message_body_empty_filter_policy_match():
    topic, queue = _setup_filter_policy_test({}, filter_policy_scope="MessageBody")

    topic.publish(
        Message='{"store": "another_corp"}',
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    to_comparable_dicts(message_attributes).should.equal(
        to_comparable_dicts([{"match": {"Type": "String", "Value": "body"}}])
    )
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"store": "another_corp"}])
    )


@mock_sqs
@mock_sns
def test_filtering_exact_string_message_body_nested():
    topic, queue = _setup_filter_policy_test(
        {"store": {"name": ["example_corp"]}}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"store": {"name": "example_corp"}}),
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"match": {"Type": "String", "Value": "body"}}])
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"store": {"name": "example_corp"}}])
    )


@mock_sqs
@mock_sns
def test_filtering_exact_string_message_body_nested_no_match():
    topic, queue = _setup_filter_policy_test(
        {"store": {"name": ["example_corp"]}}, filter_policy_scope="MessageBody"
    )

    topic.publish(
        Message=json.dumps({"store": {"name": "another_corp"}}),
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_message_body_nested_prefix():
    topic, queue = _setup_filter_policy_test(
        {"store": {"name": [{"prefix": "example_corp"}]}},
        filter_policy_scope="MessageBody",
    )

    topic.publish(
        Message=json.dumps({"store": {"name": "example_corp"}}),
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"match": {"Type": "String", "Value": "body"}}])
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    to_comparable_dicts(message_bodies).should.equal(
        to_comparable_dicts([{"store": {"name": "example_corp"}}])
    )


@mock_sqs
@mock_sns
def test_filtering_message_body_nested_prefix_no_match():
    topic, queue = _setup_filter_policy_test(
        {"store": {"name": [{"prefix": "example_corp"}]}},
        filter_policy_scope="MessageBody",
    )

    topic.publish(
        Message=json.dumps({"store": {"name": "another_corp-1"}}),
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_message_body_nested_multiple_prefix():
    topic, queue = _setup_filter_policy_test(
        {
            "Records": {
                "s3": {"object": {"key": [{"prefix": "test-"}]}},
                "eventName": [{"prefix": "ObjectCreated:"}],
            }
        },
        filter_policy_scope="MessageBody",
    )

    payload = {
        "Records": [
            {
                "eventName": "ObjectCreated:Put",
                "s3": {
                    "object": {
                        "key": "test-entry.xml",
                    }
                },
            }
        ]
    }

    topic.publish(
        Message=json.dumps(payload),
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"match": {"Type": "String", "Value": "body"}}])
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([payload])


@mock_sqs
@mock_sns
def test_filtering_message_body_nested_multiple_prefix_no_match():
    topic, queue = _setup_filter_policy_test(
        {
            "Records": {
                "s3": {"object": {"key": [{"prefix": "test-"}]}},
                "eventName": [{"prefix": "ObjectCreated:"}],
            }
        },
        filter_policy_scope="MessageBody",
    )

    payload = {
        "Records": [
            {
                "eventName": "ObjectCreated:Put",
                "s3": {
                    "object": {
                        "key": "no-match-entry.xml",
                    }
                },
            }
        ]
    }

    topic.publish(
        Message=json.dumps(payload),
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_bodies = [json.loads(m.body)["Message"] for m in messages]
    message_bodies.should.equal([])
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([])


@mock_sqs
@mock_sns
def test_filtering_message_body_nested_multiple_records_partial_match():
    topic, queue = _setup_filter_policy_test(
        {
            "Records": {
                "eventName": [{"prefix": "ObjectCreated:"}],
            }
        },
        filter_policy_scope="MessageBody",
    )

    payload = {
        "Records": [
            {
                "eventName": "ObjectCreated:Put",
            },
            {
                "eventName": "ObjectDeleted:Delete",
            },
        ]
    }

    topic.publish(
        Message=json.dumps(payload),
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"match": {"Type": "String", "Value": "body"}}])
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([payload])


@mock_sqs
@mock_sns
def test_filtering_message_body_nested_multiple_records_match():
    topic, queue = _setup_filter_policy_test(
        {
            "Records": {
                "eventName": [{"prefix": "ObjectCreated:"}],
            }
        },
        filter_policy_scope="MessageBody",
    )

    payload = {
        "Records": [
            {
                "eventName": "ObjectCreated:Put",
            },
            {
                "eventName": "ObjectCreated:Put",
            },
        ]
    }

    topic.publish(
        Message=json.dumps(payload),
        MessageAttributes={"match": {"DataType": "String", "StringValue": "body"}},
    )

    messages = queue.receive_messages(MaxNumberOfMessages=5)
    message_attributes = [json.loads(m.body)["MessageAttributes"] for m in messages]
    message_attributes.should.equal([{"match": {"Type": "String", "Value": "body"}}])
    message_bodies = [json.loads(json.loads(m.body)["Message"]) for m in messages]
    message_bodies.should.equal([payload])



import base64
import boto3
import io
import json
import pytest
import zipfile

from botocore.exceptions import ClientError
from moto import mock_lambda, mock_ec2, settings
from uuid import uuid4
from unittest import SkipTest
from .utilities import (
    get_role_name,
    get_test_zip_file_error,
    get_test_zip_file1,
    get_zip_with_multiple_files,
    get_test_zip_file2,
    get_lambda_using_environment_port,
    get_lambda_using_network_mode,
    get_test_zip_largeresponse,
)
from ..markers import requires_docker

_lambda_region = "us-west-2"


@pytest.mark.network
@mock_lambda
@requires_docker
def test_invoke_function_that_throws_error():
    conn = boto3.client("lambda", _lambda_region)
    function_name = str(uuid4())[0:6]
    conn.create_function(
        FunctionName=function_name,
        Runtime="python3.8",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_test_zip_file_error()},
    )

    failure_response = conn.invoke(
        FunctionName=function_name, Payload=json.dumps({}), LogType="Tail"
    )

    assert failure_response["FunctionError"] == "Handled"

    payload = failure_response["Payload"].read().decode("utf-8")
    payload = json.loads(payload)
    assert payload["errorType"] == "Exception"
    assert payload["errorMessage"] == "I failed!"
    assert "stackTrace" in payload

    logs = base64.b64decode(failure_response["LogResult"]).decode("utf-8")
    assert "START RequestId:" in logs
    assert "I failed!" in logs
    assert "Traceback (most recent call last):" in logs
    assert "END RequestId:" in logs


@pytest.mark.network
@pytest.mark.parametrize("invocation_type", [None, "RequestResponse"])
@pytest.mark.parametrize("key", ["FunctionName", "FunctionArn"])
@mock_lambda
@requires_docker
def test_invoke_requestresponse_function(invocation_type, key):
    conn = boto3.client("lambda", _lambda_region)
    function_name = str(uuid4())[0:6]
    fxn = conn.create_function(
        FunctionName=function_name,
        Runtime="python3.8",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_test_zip_file1()},
        Description="test lambda function",
        Timeout=3,
        MemorySize=128,
        Publish=True,
    )
    name_or_arn = fxn[key]

    # Only add invocation-type keyword-argument when provided, otherwise the request
    # fails to be validated
    kw = {}
    if invocation_type:
        kw["InvocationType"] = invocation_type

    in_data = {"msg": "So long and thanks for all the fish"}
    success_result = conn.invoke(
        FunctionName=name_or_arn, Payload=json.dumps(in_data), LogType="Tail", **kw
    )

    if "FunctionError" in success_result:
        assert False, success_result["Payload"].read().decode("utf-8")

    assert success_result["StatusCode"] == 200
    assert (
        success_result["ResponseMetadata"]["HTTPHeaders"]["content-type"]
        == "application/json"
    )
    logs = base64.b64decode(success_result["LogResult"]).decode("utf-8")

    assert "START RequestId:" in logs
    assert "custom log event" in logs
    assert "END RequestId:" in logs

    payload = success_result["Payload"].read().decode("utf-8")
    assert json.loads(payload) == in_data

    # Logs should not be returned by default, only when the LogType-param is supplied
    success_result = conn.invoke(
        FunctionName=name_or_arn, Payload=json.dumps(in_data), **kw
    )

    assert success_result["StatusCode"] == 200
    assert (
        success_result["ResponseMetadata"]["HTTPHeaders"]["content-type"]
        == "application/json"
    )
    assert "LogResult" not in success_result


@pytest.mark.network
@mock_lambda
@requires_docker
def test_invoke_event_function():
    conn = boto3.client("lambda", _lambda_region)
    function_name = str(uuid4())[0:6]
    conn.create_function(
        FunctionName=function_name,
        Runtime="python3.9",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_test_zip_file1()},
        Description="test lambda function",
        Timeout=3,
        MemorySize=128,
        Publish=True,
    )

    with pytest.raises(ClientError):
        conn.invoke(FunctionName="notAFunction", InvocationType="Event", Payload="{}")

    in_data = {"msg": "So long and thanks for all the fish"}
    success_result = conn.invoke(
        FunctionName=function_name, InvocationType="Event", Payload=json.dumps(in_data)
    )
    assert success_result["StatusCode"] == 202
    assert json.loads(success_result["Payload"].read().decode("utf-8")) == in_data


@pytest.mark.network
@mock_lambda
def test_invoke_lambda_using_environment_port():
    if not settings.TEST_SERVER_MODE:
        raise SkipTest("Can only test environment variables in server mode")
    conn = boto3.client("lambda", _lambda_region)
    function_name = str(uuid4())[0:6]
    conn.create_function(
        FunctionName=function_name,
        Runtime="python3.7",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_lambda_using_environment_port()},
    )

    success_result = conn.invoke(
        FunctionName=function_name, InvocationType="Event", Payload="{}"
    )

    assert success_result["StatusCode"] == 202
    response = success_result["Payload"].read()
    response = json.loads(response.decode("utf-8"))

    functions = response["functions"]
    function_names = [f["FunctionName"] for f in functions]
    assert function_name in function_names

    # Host matches the full URL, so one of:
    # http://host.docker.internal:5000
    # http://172.0.2.1:5000
    # http://172.0.1.1:4555
    assert "http://" in response["host"]


@pytest.mark.network
@mock_lambda
def test_invoke_lambda_using_networkmode():
    """
    Special use case - verify that Lambda can send a request to 'http://localhost'
    This is only possible when the `network_mode` is set to host in the Docker args
    Test is only run in our CI (for now)
    """
    if not settings.moto_network_mode():
        raise SkipTest("Can only test this when NETWORK_MODE is specified")
    conn = boto3.client("lambda", _lambda_region)
    function_name = str(uuid4())[0:6]
    conn.create_function(
        FunctionName=function_name,
        Runtime="python3.7",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_lambda_using_network_mode()},
    )

    success_result = conn.invoke(
        FunctionName=function_name, InvocationType="Event", Payload="{}"
    )

    response = success_result["Payload"].read()
    functions = json.loads(response.decode("utf-8"))["response"]
    function_names = [f["FunctionName"] for f in functions]
    assert function_name in function_names


@pytest.mark.network
@mock_lambda
@requires_docker
def test_invoke_function_with_multiple_files_in_zip():
    conn = boto3.client("lambda", _lambda_region)
    function_name = str(uuid4())[0:6]
    conn.create_function(
        FunctionName=function_name,
        Runtime="python3.7",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_zip_with_multiple_files()},
        Description="test lambda function",
        Timeout=3,
        MemorySize=128,
        Publish=True,
    )

    in_data = {"msg": "So long and thanks for: "}
    success_result = conn.invoke(
        FunctionName=function_name, InvocationType="Event", Payload=json.dumps(in_data)
    )
    assert json.loads(success_result["Payload"].read().decode("utf-8")) == {
        "msg": "So long and thanks for: stuff"
    }


@pytest.mark.network
@mock_lambda
@requires_docker
def test_invoke_dryrun_function():
    conn = boto3.client("lambda", _lambda_region)
    function_name = str(uuid4())[0:6]
    conn.create_function(
        FunctionName=function_name,
        Runtime="python3.8",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_test_zip_file1()},
        Description="test lambda function",
        Timeout=3,
        MemorySize=128,
        Publish=True,
    )

    with pytest.raises(ClientError):
        conn.invoke(FunctionName="notAFunction", InvocationType="Event", Payload="{}")

    in_data = {"msg": "So long and thanks for all the fish"}
    success_result = conn.invoke(
        FunctionName=function_name, InvocationType="DryRun", Payload=json.dumps(in_data)
    )
    assert success_result["StatusCode"] == 204


if settings.TEST_SERVER_MODE:

    @mock_ec2
    @mock_lambda
    def test_invoke_function_get_ec2_volume():
        conn = boto3.resource("ec2", _lambda_region)
        vol = conn.create_volume(Size=99, AvailabilityZone=_lambda_region)
        vol = conn.Volume(vol.id)

        conn = boto3.client("lambda", _lambda_region)
        function_name = str(uuid4())[0:6]
        conn.create_function(
            FunctionName=function_name,
            Runtime="python3.7",
            Role=get_role_name(),
            Handler="lambda_function.lambda_handler",
            Code={"ZipFile": get_test_zip_file2()},
            Description="test lambda function",
            Timeout=3,
            MemorySize=128,
            Publish=True,
        )

        in_data = {"volume_id": vol.id}
        result = conn.invoke(
            FunctionName=function_name,
            InvocationType="RequestResponse",
            Payload=json.dumps(in_data),
        )
        assert result["StatusCode"] == 200
        actual_payload = json.loads(result["Payload"].read().decode("utf-8"))
        expected_payload = {"id": vol.id, "state": vol.state, "size": vol.size}
        assert actual_payload == expected_payload


@pytest.mark.network
@mock_lambda
@requires_docker
def test_invoke_lambda_error():
    lambda_fx = """
def lambda_handler(event, context):
    raise Exception('failsauce')
    """
    zip_output = io.BytesIO()
    zip_file = zipfile.ZipFile(zip_output, "w", zipfile.ZIP_DEFLATED)
    zip_file.writestr("lambda_function.py", lambda_fx)
    zip_file.close()
    zip_output.seek(0)

    client = boto3.client("lambda", region_name="us-east-1")
    client.create_function(
        FunctionName="test-lambda-fx",
        Runtime="python3.8",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Description="test lambda function",
        Timeout=3,
        MemorySize=128,
        Publish=True,
        Code={"ZipFile": zip_output.read()},
    )

    result = client.invoke(
        FunctionName="test-lambda-fx", InvocationType="RequestResponse", LogType="Tail"
    )

    assert "FunctionError" in result
    assert result["FunctionError"] == "Handled"


@pytest.mark.network
@pytest.mark.parametrize("key", ["FunctionName", "FunctionArn"])
@mock_lambda
@requires_docker
def test_invoke_async_function(key):
    conn = boto3.client("lambda", _lambda_region)
    function_name = str(uuid4())[0:6]
    fxn = conn.create_function(
        FunctionName=function_name,
        Runtime="python3.8",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_test_zip_file1()},
        Description="test lambda function",
        Timeout=3,
        MemorySize=128,
        Publish=True,
    )
    name_or_arn = fxn[key]

    success_result = conn.invoke_async(
        FunctionName=name_or_arn, InvokeArgs=json.dumps({"test": "event"})
    )

    assert success_result["Status"] == 202


@pytest.mark.network
@mock_lambda
@requires_docker
def test_invoke_function_large_response():
    # AWS Lambda should only return bodies smaller than 6 MB
    conn = boto3.client("lambda", _lambda_region)
    fxn = conn.create_function(
        FunctionName=str(uuid4())[0:6],
        Runtime="python3.7",
        Role=get_role_name(),
        Handler="lambda_function.lambda_handler",
        Code={"ZipFile": get_test_zip_largeresponse()},
        Description="test lambda function",
        Timeout=3,
        MemorySize=128,
        Publish=True,
    )

    resp = conn.invoke(FunctionName=fxn["FunctionArn"])
    assert resp["FunctionError"] == "Unhandled"
    payload = resp["Payload"].read().decode("utf-8")
    payload = json.loads(payload)
    assert payload == {
        "errorMessage": "Response payload size exceeded maximum allowed payload size (6291556 bytes).",
        "errorType": "Function.ResponseSizeTooLarge",
    }

    # Absolutely fine when invoking async
    resp = conn.invoke(FunctionName=fxn["FunctionArn"], InvocationType="Event")
    assert "FunctionError" not in resp



from moto.core.exceptions import JsonRESTError
from typing import Optional


class AliasExistsException(JsonRESTError):
    def __init__(self) -> None:
        super().__init__(
            "AliasExistsException", "An account with the given email already exists."
        )


class ResourceNotFoundError(JsonRESTError):
    def __init__(self, message: Optional[str]):
        super().__init__(error_type="ResourceNotFoundException", message=message or "")


class UserNotFoundError(JsonRESTError):
    def __init__(self, message: str):
        super().__init__(error_type="UserNotFoundException", message=message)


class UsernameExistsException(JsonRESTError):
    def __init__(self, message: str):
        super().__init__(error_type="UsernameExistsException", message=message)


class GroupExistsException(JsonRESTError):
    def __init__(self, message: str):
        super().__init__(error_type="GroupExistsException", message=message)


class NotAuthorizedError(JsonRESTError):
    def __init__(self, message: Optional[str]):
        super().__init__(error_type="NotAuthorizedException", message=message or "")


class UserNotConfirmedException(JsonRESTError):
    def __init__(self, message: str):
        super().__init__(error_type="UserNotConfirmedException", message=message)


class ExpiredCodeException(JsonRESTError):
    def __init__(self, message: str):
        super().__init__(error_type="ExpiredCodeException", message=message)


class InvalidParameterException(JsonRESTError):
    def __init__(self, msg: Optional[str] = None):
        self.code = 400
        super().__init__(
            "InvalidParameterException", msg or "A parameter is specified incorrectly."
        )


class InvalidPasswordException(JsonRESTError):
    def __init__(self) -> None:
        super().__init__(
            error_type="InvalidPasswordException",
            message="The provided password does not confirm to the configured password policy",
        )



import boto3
from botocore.exceptions import ClientError
import datetime
import sure  # noqa # pylint: disable=unused-import
import pytest

from moto import mock_sagemaker
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID

FAKE_ROLE_ARN = f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole"
TEST_REGION_NAME = "us-east-1"


class MyTrainingJobModel(object):
    def __init__(
        self,
        training_job_name,
        role_arn,
        container=None,
        bucket=None,
        prefix=None,
        algorithm_specification=None,
        resource_config=None,
        input_data_config=None,
        output_data_config=None,
        hyper_parameters=None,
        stopping_condition=None,
    ):
        self.training_job_name = training_job_name
        self.role_arn = role_arn
        self.container = (
            container or "382416733822.dkr.ecr.us-east-1.amazonaws.com/linear-learner:1"
        )
        self.bucket = bucket or "my-bucket"
        self.prefix = prefix or "sagemaker/DEMO-breast-cancer-prediction/"
        self.algorithm_specification = algorithm_specification or {
            "TrainingImage": self.container,
            "TrainingInputMode": "File",
        }
        self.resource_config = resource_config or {
            "InstanceCount": 1,
            "InstanceType": "ml.c4.2xlarge",
            "VolumeSizeInGB": 10,
        }
        self.input_data_config = input_data_config or [
            {
                "ChannelName": "train",
                "DataSource": {
                    "S3DataSource": {
                        "S3DataType": "S3Prefix",
                        "S3Uri": f"s3://{self.bucket}/{self.prefix}/train/",
                        "S3DataDistributionType": "ShardedByS3Key",
                    }
                },
                "CompressionType": "None",
                "RecordWrapperType": "None",
            },
            {
                "ChannelName": "validation",
                "DataSource": {
                    "S3DataSource": {
                        "S3DataType": "S3Prefix",
                        "S3Uri": f"s3://{self.bucket}/{self.prefix}/validation/",
                        "S3DataDistributionType": "FullyReplicated",
                    }
                },
                "CompressionType": "None",
                "RecordWrapperType": "None",
            },
        ]
        self.output_data_config = output_data_config or {
            "S3OutputPath": f"s3://{self.bucket}/{self.prefix}/"
        }
        self.hyper_parameters = hyper_parameters or {
            "feature_dim": "30",
            "mini_batch_size": "100",
            "predictor_type": "regressor",
            "epochs": "10",
            "num_models": "32",
            "loss": "absolute_loss",
        }

        self.stopping_condition = stopping_condition or {"MaxRuntimeInSeconds": 60 * 60}

    def save(self):
        sagemaker = boto3.client("sagemaker", region_name=TEST_REGION_NAME)

        params = {
            "RoleArn": self.role_arn,
            "TrainingJobName": self.training_job_name,
            "AlgorithmSpecification": self.algorithm_specification,
            "ResourceConfig": self.resource_config,
            "InputDataConfig": self.input_data_config,
            "OutputDataConfig": self.output_data_config,
            "HyperParameters": self.hyper_parameters,
            "StoppingCondition": self.stopping_condition,
        }
        return sagemaker.create_training_job(**params)


@mock_sagemaker
def test_create_training_job():
    sagemaker = boto3.client("sagemaker", region_name=TEST_REGION_NAME)

    training_job_name = "MyTrainingJob"
    role_arn = f"arn:aws:iam::{ACCOUNT_ID}:role/FakeRole"
    container = "382416733822.dkr.ecr.us-east-1.amazonaws.com/linear-learner:1"
    bucket = "my-bucket"
    prefix = "sagemaker/DEMO-breast-cancer-prediction/"
    algorithm_specification = {
        "TrainingImage": container,
        "TrainingInputMode": "File",
    }
    resource_config = {
        "InstanceCount": 1,
        "InstanceType": "ml.c4.2xlarge",
        "VolumeSizeInGB": 10,
    }
    input_data_config = [
        {
            "ChannelName": "train",
            "DataSource": {
                "S3DataSource": {
                    "S3DataType": "S3Prefix",
                    "S3Uri": f"s3://{bucket}/{prefix}/train/",
                    "S3DataDistributionType": "ShardedByS3Key",
                }
            },
            "CompressionType": "None",
            "RecordWrapperType": "None",
        },
        {
            "ChannelName": "validation",
            "DataSource": {
                "S3DataSource": {
                    "S3DataType": "S3Prefix",
                    "S3Uri": f"s3://{bucket}/{prefix}/validation/",
                    "S3DataDistributionType": "FullyReplicated",
                }
            },
            "CompressionType": "None",
            "RecordWrapperType": "None",
        },
    ]
    output_data_config = {"S3OutputPath": f"s3://{bucket}/{prefix}/"}
    hyper_parameters = {
        "feature_dim": "30",
        "mini_batch_size": "100",
        "predictor_type": "regressor",
        "epochs": "10",
        "num_models": "32",
        "loss": "absolute_loss",
    }
    stopping_condition = {"MaxRuntimeInSeconds": 60 * 60}

    job = MyTrainingJobModel(
        training_job_name,
        role_arn,
        container=container,
        bucket=bucket,
        prefix=prefix,
        algorithm_specification=algorithm_specification,
        resource_config=resource_config,
        input_data_config=input_data_config,
        output_data_config=output_data_config,
        hyper_parameters=hyper_parameters,
        stopping_condition=stopping_condition,
    )
    resp = job.save()
    resp["TrainingJobArn"].should.match(
        rf"^arn:aws:sagemaker:.*:.*:training-job/{training_job_name}$"
    )

    resp = sagemaker.describe_training_job(TrainingJobName=training_job_name)
    resp["TrainingJobName"].should.equal(training_job_name)
    resp["TrainingJobArn"].should.match(
        rf"^arn:aws:sagemaker:.*:.*:training-job/{training_job_name}$"
    )
    assert resp["ModelArtifacts"]["S3ModelArtifacts"].startswith(
        output_data_config["S3OutputPath"]
    )
    assert training_job_name in (resp["ModelArtifacts"]["S3ModelArtifacts"])
    assert resp["ModelArtifacts"]["S3ModelArtifacts"].endswith("output/model.tar.gz")
    assert resp["TrainingJobStatus"] == "Completed"
    assert resp["SecondaryStatus"] == "Completed"
    assert resp["HyperParameters"] == hyper_parameters
    assert (
        resp["AlgorithmSpecification"]["TrainingImage"]
        == algorithm_specification["TrainingImage"]
    )
    assert (
        resp["AlgorithmSpecification"]["TrainingInputMode"]
        == algorithm_specification["TrainingInputMode"]
    )
    assert "MetricDefinitions" in resp["AlgorithmSpecification"]
    assert "Name" in resp["AlgorithmSpecification"]["MetricDefinitions"][0]
    assert "Regex" in resp["AlgorithmSpecification"]["MetricDefinitions"][0]
    assert resp["RoleArn"] == role_arn
    assert resp["InputDataConfig"] == input_data_config
    assert resp["OutputDataConfig"] == output_data_config
    assert resp["ResourceConfig"] == resource_config
    assert resp["StoppingCondition"] == stopping_condition
    assert isinstance(resp["CreationTime"], datetime.datetime)
    assert isinstance(resp["TrainingStartTime"], datetime.datetime)
    assert isinstance(resp["TrainingEndTime"], datetime.datetime)
    assert isinstance(resp["LastModifiedTime"], datetime.datetime)
    assert "SecondaryStatusTransitions" in resp
    assert "Status" in resp["SecondaryStatusTransitions"][0]
    assert "StartTime" in resp["SecondaryStatusTransitions"][0]
    assert "EndTime" in resp["SecondaryStatusTransitions"][0]
    assert "StatusMessage" in resp["SecondaryStatusTransitions"][0]
    assert "FinalMetricDataList" in resp
    assert "MetricName" in resp["FinalMetricDataList"][0]
    assert "Value" in resp["FinalMetricDataList"][0]
    assert "Timestamp" in resp["FinalMetricDataList"][0]

    pass


@mock_sagemaker
def test_list_training_jobs():
    client = boto3.client("sagemaker", region_name="us-east-1")
    name = "blah"
    arn = "arn:aws:sagemaker:us-east-1:000000000000:x-x/foobar"
    test_training_job = MyTrainingJobModel(training_job_name=name, role_arn=arn)
    test_training_job.save()
    training_jobs = client.list_training_jobs()
    assert len(training_jobs["TrainingJobSummaries"]).should.equal(1)
    assert training_jobs["TrainingJobSummaries"][0]["TrainingJobName"].should.equal(
        name
    )

    assert training_jobs["TrainingJobSummaries"][0]["TrainingJobArn"].should.match(
        rf"^arn:aws:sagemaker:.*:.*:training-job/{name}$"
    )
    assert training_jobs.get("NextToken") is None


@mock_sagemaker
def test_list_training_jobs_multiple():
    client = boto3.client("sagemaker", region_name="us-east-1")
    name_job_1 = "blah"
    arn_job_1 = "arn:aws:sagemaker:us-east-1:000000000000:x-x/foobar"
    test_training_job_1 = MyTrainingJobModel(
        training_job_name=name_job_1, role_arn=arn_job_1
    )
    test_training_job_1.save()

    name_job_2 = "blah2"
    arn_job_2 = "arn:aws:sagemaker:us-east-1:000000000000:x-x/foobar2"
    test_training_job_2 = MyTrainingJobModel(
        training_job_name=name_job_2, role_arn=arn_job_2
    )
    test_training_job_2.save()
    training_jobs_limit = client.list_training_jobs(MaxResults=1)
    assert len(training_jobs_limit["TrainingJobSummaries"]).should.equal(1)

    training_jobs = client.list_training_jobs()
    assert len(training_jobs["TrainingJobSummaries"]).should.equal(2)
    assert training_jobs.get("NextToken").should.be.none


@mock_sagemaker
def test_list_training_jobs_none():
    client = boto3.client("sagemaker", region_name="us-east-1")
    training_jobs = client.list_training_jobs()
    assert len(training_jobs["TrainingJobSummaries"]).should.equal(0)


@mock_sagemaker
def test_list_training_jobs_should_validate_input():
    client = boto3.client("sagemaker", region_name="us-east-1")
    junk_status_equals = "blah"
    with pytest.raises(ClientError) as ex:
        client.list_training_jobs(StatusEquals=junk_status_equals)
    expected_error = f"1 validation errors detected: Value '{junk_status_equals}' at 'statusEquals' failed to satisfy constraint: Member must satisfy enum value set: ['Completed', 'Stopped', 'InProgress', 'Stopping', 'Failed']"
    assert ex.value.response["Error"]["Code"] == "ValidationException"
    assert ex.value.response["Error"]["Message"] == expected_error

    junk_next_token = "asdf"
    with pytest.raises(ClientError) as ex:
        client.list_training_jobs(NextToken=junk_next_token)
    assert ex.value.response["Error"]["Code"] == "ValidationException"
    assert (
        ex.value.response["Error"]["Message"]
        == 'Invalid pagination token because "{0}".'
    )


@mock_sagemaker
def test_list_training_jobs_with_name_filters():
    client = boto3.client("sagemaker", region_name="us-east-1")
    for i in range(5):
        name = f"xgboost-{i}"
        arn = f"arn:aws:sagemaker:us-east-1:000000000000:x-x/foobar-{i}"
        MyTrainingJobModel(training_job_name=name, role_arn=arn).save()
    for i in range(5):
        name = f"vgg-{i}"
        arn = f"arn:aws:sagemaker:us-east-1:000000000000:x-x/barfoo-{i}"
        MyTrainingJobModel(training_job_name=name, role_arn=arn).save()
    xgboost_training_jobs = client.list_training_jobs(NameContains="xgboost")
    assert len(xgboost_training_jobs["TrainingJobSummaries"]).should.equal(5)

    training_jobs_with_2 = client.list_training_jobs(NameContains="2")
    assert len(training_jobs_with_2["TrainingJobSummaries"]).should.equal(2)


@mock_sagemaker
def test_list_training_jobs_paginated():
    client = boto3.client("sagemaker", region_name="us-east-1")
    for i in range(5):
        name = f"xgboost-{i}"
        arn = f"arn:aws:sagemaker:us-east-1:000000000000:x-x/foobar-{i}"
        MyTrainingJobModel(training_job_name=name, role_arn=arn).save()
    xgboost_training_job_1 = client.list_training_jobs(
        NameContains="xgboost", MaxResults=1
    )
    assert len(xgboost_training_job_1["TrainingJobSummaries"]).should.equal(1)
    assert xgboost_training_job_1["TrainingJobSummaries"][0][
        "TrainingJobName"
    ].should.equal("xgboost-0")
    assert xgboost_training_job_1.get("NextToken").should_not.be.none

    xgboost_training_job_next = client.list_training_jobs(
        NameContains="xgboost",
        MaxResults=1,
        NextToken=xgboost_training_job_1.get("NextToken"),
    )
    assert len(xgboost_training_job_next["TrainingJobSummaries"]).should.equal(1)
    assert xgboost_training_job_next["TrainingJobSummaries"][0][
        "TrainingJobName"
    ].should.equal("xgboost-1")
    assert xgboost_training_job_next.get("NextToken").should_not.be.none


@mock_sagemaker
def test_list_training_jobs_paginated_with_target_in_middle():
    client = boto3.client("sagemaker", region_name="us-east-1")
    for i in range(5):
        name = f"xgboost-{i}"
        arn = f"arn:aws:sagemaker:us-east-1:000000000000:x-x/foobar-{i}"
        MyTrainingJobModel(training_job_name=name, role_arn=arn).save()
    for i in range(5):
        name = f"vgg-{i}"
        arn = f"arn:aws:sagemaker:us-east-1:000000000000:x-x/barfoo-{i}"
        MyTrainingJobModel(training_job_name=name, role_arn=arn).save()

    vgg_training_job_1 = client.list_training_jobs(NameContains="vgg", MaxResults=1)
    assert len(vgg_training_job_1["TrainingJobSummaries"]).should.equal(0)
    assert vgg_training_job_1.get("NextToken").should_not.be.none

    vgg_training_job_6 = client.list_training_jobs(NameContains="vgg", MaxResults=6)

    assert len(vgg_training_job_6["TrainingJobSummaries"]).should.equal(1)
    assert vgg_training_job_6["TrainingJobSummaries"][0][
        "TrainingJobName"
    ].should.equal("vgg-0")
    assert vgg_training_job_6.get("NextToken").should_not.be.none

    vgg_training_job_10 = client.list_training_jobs(NameContains="vgg", MaxResults=10)

    assert len(vgg_training_job_10["TrainingJobSummaries"]).should.equal(5)
    assert vgg_training_job_10["TrainingJobSummaries"][-1][
        "TrainingJobName"
    ].should.equal("vgg-4")
    assert vgg_training_job_10.get("NextToken").should.be.none


@mock_sagemaker
def test_list_training_jobs_paginated_with_fragmented_targets():
    client = boto3.client("sagemaker", region_name="us-east-1")
    for i in range(5):
        name = f"xgboost-{i}"
        arn = f"arn:aws:sagemaker:us-east-1:000000000000:x-x/foobar-{i}"
        MyTrainingJobModel(training_job_name=name, role_arn=arn).save()
    for i in range(5):
        name = f"vgg-{i}"
        arn = f"arn:aws:sagemaker:us-east-1:000000000000:x-x/barfoo-{i}"
        MyTrainingJobModel(training_job_name=name, role_arn=arn).save()

    training_jobs_with_2 = client.list_training_jobs(NameContains="2", MaxResults=8)
    assert len(training_jobs_with_2["TrainingJobSummaries"]).should.equal(2)
    assert training_jobs_with_2.get("NextToken").should_not.be.none

    training_jobs_with_2_next = client.list_training_jobs(
        NameContains="2", MaxResults=1, NextToken=training_jobs_with_2.get("NextToken")
    )
    assert len(training_jobs_with_2_next["TrainingJobSummaries"]).should.equal(0)
    assert training_jobs_with_2_next.get("NextToken").should_not.be.none

    training_jobs_with_2_next_next = client.list_training_jobs(
        NameContains="2",
        MaxResults=1,
        NextToken=training_jobs_with_2_next.get("NextToken"),
    )
    assert len(training_jobs_with_2_next_next["TrainingJobSummaries"]).should.equal(0)
    assert training_jobs_with_2_next_next.get("NextToken").should.be.none


@mock_sagemaker
def test_add_tags_to_training_job():
    client = boto3.client("sagemaker", region_name=TEST_REGION_NAME)
    name = "blah"
    resource_arn = f"arn:aws:sagemaker:us-east-1:000000000000:training-job/{name}"
    test_training_job = MyTrainingJobModel(
        training_job_name=name, role_arn=resource_arn
    )
    test_training_job.save()

    tags = [
        {"Key": "myKey", "Value": "myValue"},
    ]
    response = client.add_tags(ResourceArn=resource_arn, Tags=tags)
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = client.list_tags(ResourceArn=resource_arn)
    assert response["Tags"] == tags


@mock_sagemaker
def test_delete_tags_from_training_job():
    client = boto3.client("sagemaker", region_name=TEST_REGION_NAME)
    name = "blah"
    resource_arn = f"arn:aws:sagemaker:us-east-1:000000000000:training-job/{name}"
    test_training_job = MyTrainingJobModel(
        training_job_name=name, role_arn=resource_arn
    )
    test_training_job.save()

    tags = [
        {"Key": "myKey", "Value": "myValue"},
    ]
    response = client.add_tags(ResourceArn=resource_arn, Tags=tags)
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    tag_keys = [tag["Key"] for tag in tags]
    response = client.delete_tags(ResourceArn=resource_arn, TagKeys=tag_keys)
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    response = client.list_tags(ResourceArn=resource_arn)
    assert response["Tags"] == []


@mock_sagemaker
def test_describe_unknown_training_job():
    client = boto3.client("sagemaker", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.describe_training_job(TrainingJobName="unknown")
    err = exc.value.response["Error"]
    err["Code"].should.equal("ValidationException")
    err["Message"].should.equal(
        f"Could not find training job 'arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:training-job/unknown'."
    )



import boto3
import pytest

from botocore.exceptions import ClientError

from moto import mock_sagemaker

TEST_REGION_NAME = "us-east-1"


@pytest.fixture(name="sagemaker_client")
def fixture_sagemaker_client():
    with mock_sagemaker():
        yield boto3.client("sagemaker", region_name=TEST_REGION_NAME)


def test_search(sagemaker_client):
    experiment_name = "experiment_name"
    trial_component_name = "trial_component_name"
    trial_name = "trial_name"
    _set_up_trial_component(
        sagemaker_client,
        experiment_name=experiment_name,
        trial_component_name=trial_component_name,
        trial_name=trial_name,
    )

    resp = sagemaker_client.search(Resource="ExperimentTrialComponent")
    assert len(resp["Results"]) == 2

    resp = sagemaker_client.describe_trial_component(
        TrialComponentName=trial_component_name
    )
    trial_component_arn = resp["TrialComponentArn"]

    tags = [{"Key": "key-name", "Value": "some-value"}]
    sagemaker_client.add_tags(ResourceArn=trial_component_arn, Tags=tags)

    resp = sagemaker_client.search(
        Resource="ExperimentTrialComponent",
        SearchExpression={
            "Filters": [
                {"Name": "Tags.key-name", "Operator": "Equals", "Value": "some-value"}
            ]
        },
    )

    assert len(resp["Results"]) == 1
    assert (
        resp["Results"][0]["TrialComponent"]["TrialComponentName"]
        == trial_component_name
    )

    resp = sagemaker_client.search(Resource="Experiment")
    assert len(resp["Results"]) == 1
    assert resp["Results"][0]["Experiment"]["ExperimentName"] == experiment_name

    resp = sagemaker_client.search(Resource="ExperimentTrial")
    assert len(resp["Results"]) == 1
    assert resp["Results"][0]["Trial"]["TrialName"] == trial_name


def test_search_trial_component_with_experiment_name(sagemaker_client):
    experiment_name = "experiment_name"
    trial_component_name = "trial_component_name"
    _set_up_trial_component(
        sagemaker_client,
        experiment_name=experiment_name,
        trial_component_name=trial_component_name,
    )

    resp = sagemaker_client.search(Resource="ExperimentTrialComponent")
    assert len(resp["Results"]) == 2

    resp = sagemaker_client.describe_trial_component(
        TrialComponentName=trial_component_name
    )
    trial_component_arn = resp["TrialComponentArn"]

    tags = [{"Key": "key-name", "Value": "some-value"}]
    sagemaker_client.add_tags(ResourceArn=trial_component_arn, Tags=tags)

    with pytest.raises(ClientError) as ex:
        sagemaker_client.search(
            Resource="ExperimentTrialComponent",
            SearchExpression={
                "Filters": [
                    {
                        "Name": "ExperimentName",
                        "Operator": "Equals",
                        "Value": experiment_name,
                    }
                ]
            },
        )

    ex.value.response["Error"]["Code"].should.equal("ValidationException")
    ex.value.response["Error"]["Message"].should.equal(
        "Unknown property name: ExperimentName"
    )
    ex.value.response["ResponseMetadata"]["HTTPStatusCode"].should.equal(400)


def _set_up_trial_component(
    sagemaker_client,
    experiment_name="some-experiment-name",
    trial_component_name="some-trial-component-name",
    trial_name="some-trial-name",
    another_trial_component_name="another-trial-component-name",
):
    sagemaker_client.create_experiment(ExperimentName=experiment_name)
    sagemaker_client.create_trial(ExperimentName=experiment_name, TrialName=trial_name)
    sagemaker_client.create_trial_component(TrialComponentName=trial_component_name)
    sagemaker_client.create_trial_component(
        TrialComponentName=another_trial_component_name
    )



import boto3

from moto import mock_rds


@mock_rds
def test_describe_db_cluster_parameters():
    client = boto3.client("rds", "us-east-2")

    resp = client.describe_db_cluster_parameters(DBClusterParameterGroupName="group")
    assert resp["Parameters"] == []



from moto.ses.template import parse_template


def test_template_without_args():
    assert parse_template("some template", template_data={}) == "some template"


def test_template_with_simple_arg():
    template = "Dear {{name}},"
    assert parse_template(template, template_data={"name": "John"}) == "Dear John,"


def test_template_with_foreach():
    template = "List:{{#each l}} - {{obj}}{{/each}} and other things"
    kwargs = {"l": [{"obj": "it1"}, {"obj": "it2"}]}
    assert parse_template(template, kwargs) == "List: - it1 - it2 and other things"


def test_template_with_multiple_foreach():
    template = (
        "{{#each l}} - {{obj}}{{/each}} and list 2 {{#each l2}} {{o1}} {{o2}} {{/each}}"
    )
    kwargs = {
        "l": [{"obj": "it1"}, {"obj": "it2"}],
        "l2": [{"o1": "ob1", "o2": "ob2"}, {"o1": "oc1", "o2": "oc2"}],
    }
    assert (
        parse_template(template, kwargs) == " - it1 - it2 and list 2  ob1 ob2  oc1 oc2 "
    )


def test_template_with_single_curly_brace():
    template = "Dear {{name}} my {brace} is fine."
    assert parse_template(template, template_data={"name": "John"}) == (
        "Dear John my {brace} is fine."
    )



import json

import moto.server as server
from moto import mock_budgets


@mock_budgets
def test_budgets_describe_budgets():
    backend = server.create_backend_app("budgets")
    test_client = backend.test_client()

    headers = {"X-Amz-Target": "AWSBudgetServiceGateway.DescribeBudgets"}
    resp = test_client.post("/", headers=headers, json={})
    assert resp.status_code == 200
    assert json.loads(resp.data) == {"Budgets": [], "nextToken": None}



from moto.core.exceptions import RESTError


class AutoscalingClientError(RESTError):
    code = 400


class ResourceContentionError(RESTError):
    code = 500

    def __init__(self) -> None:
        super().__init__(
            "ResourceContentionError",
            "You already have a pending update to an Auto Scaling resource (for example, a group, instance, or load balancer).",
        )


class InvalidInstanceError(AutoscalingClientError):
    def __init__(self, instance_id: str):
        super().__init__("ValidationError", f"Instance [{instance_id}] is invalid.")


class ValidationError(AutoscalingClientError):
    def __init__(self, message: str):
        super().__init__("ValidationError", message)



import boto3

from moto import mock_iot


@mock_iot
def test_create_thing_type():
    client = boto3.client("iot", region_name="ap-northeast-1")
    type_name = "my-type-name"

    thing_type = client.create_thing_type(thingTypeName=type_name)
    assert thing_type["thingTypeName"] == type_name
    assert type_name in thing_type["thingTypeArn"]


@mock_iot
def test_describe_thing_type():
    client = boto3.client("iot", region_name="ap-northeast-1")
    type_name = "my-type-name"

    client.create_thing_type(thingTypeName=type_name)

    thing_type = client.describe_thing_type(thingTypeName=type_name)
    assert thing_type["thingTypeName"] == type_name
    assert "thingTypeProperties" in thing_type
    assert "thingTypeMetadata" in thing_type
    assert type_name in thing_type["thingTypeArn"]


@mock_iot
def test_list_thing_types():
    client = boto3.client("iot", region_name="ap-northeast-1")

    for i in range(0, 100):
        client.create_thing_type(thingTypeName=str(i + 1))

    thing_types = client.list_thing_types()
    assert "nextToken" in thing_types
    assert len(thing_types["thingTypes"]) == 50
    assert thing_types["thingTypes"][0]["thingTypeName"] == "1"
    assert thing_types["thingTypes"][-1]["thingTypeName"] == "50"

    thing_types = client.list_thing_types(nextToken=thing_types["nextToken"])
    assert len(thing_types["thingTypes"]) == 50
    assert "nextToken" not in thing_types
    assert thing_types["thingTypes"][0]["thingTypeName"] == "51"
    assert thing_types["thingTypes"][-1]["thingTypeName"] == "100"


@mock_iot
def test_list_thing_types_with_typename_filter():
    client = boto3.client("iot", region_name="ap-northeast-1")

    client.create_thing_type(thingTypeName="thing")
    client.create_thing_type(thingTypeName="thingType")
    client.create_thing_type(thingTypeName="thingTypeName")
    client.create_thing_type(thingTypeName="thingTypeNameGroup")
    client.create_thing_type(thingTypeName="shouldNotFind")
    client.create_thing_type(thingTypeName="find me it shall not")

    thing_types = client.list_thing_types(thingTypeName="thing")
    assert "nextToken" not in thing_types
    assert len(thing_types["thingTypes"]) == 4
    assert thing_types["thingTypes"][0]["thingTypeName"] == "thing"
    assert thing_types["thingTypes"][-1]["thingTypeName"] == "thingTypeNameGroup"

    thing_types = client.list_thing_types(thingTypeName="thingTypeName")
    assert "nextToken" not in thing_types
    assert len(thing_types["thingTypes"]) == 2
    assert thing_types["thingTypes"][0]["thingTypeName"] == "thingTypeName"
    assert thing_types["thingTypes"][-1]["thingTypeName"] == "thingTypeNameGroup"


@mock_iot
def test_delete_thing_type():
    client = boto3.client("iot", region_name="ap-northeast-1")
    type_name = "my-type-name"

    client.create_thing_type(thingTypeName=type_name)

    # delete thing type
    client.delete_thing_type(thingTypeName=type_name)
    res = client.list_thing_types()
    assert len(res["thingTypes"]) == 0



"""Unit tests for cloudtrail-supported APIs."""
import boto3
import pytest

from botocore.exceptions import ClientError
from datetime import datetime
from moto import mock_cloudtrail, mock_s3, mock_sns
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from uuid import uuid4


@mock_s3
@mock_cloudtrail
def test_create_trail_without_bucket():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.create_trail(
            Name="mytrailname", S3BucketName="specificweirdbucketthatdoesnotexist"
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "S3BucketDoesNotExistException"
    assert (
        err["Message"]
        == "S3 bucket specificweirdbucketthatdoesnotexist does not exist!"
    )


@pytest.mark.parametrize(
    "name,message",
    [
        (
            "a",
            "Trail name too short. Minimum allowed length: 3 characters. Specified name length: 1 characters.",
        ),
        (
            "aa",
            "Trail name too short. Minimum allowed length: 3 characters. Specified name length: 2 characters.",
        ),
        (
            "a" * 129,
            "Trail name too long. Maximum allowed length: 128 characters. Specified name length: 129 characters.",
        ),
        ("trail!", "Trail name must ends with a letter or number."),
        (
            "my#trail",
            "Trail name or ARN can only contain uppercase letters, lowercase letters, numbers, periods (.), hyphens (-), and underscores (_).",
        ),
        ("-trail", "Trail name must starts with a letter or number."),
    ],
)
@mock_cloudtrail
def test_create_trail_invalid_name(name, message):
    client = boto3.client("cloudtrail", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.create_trail(
            Name=name, S3BucketName="specificweirdbucketthatdoesnotexist"
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidTrailNameException"
    assert err["Message"] == message


@mock_cloudtrail
@mock_s3
def test_create_trail_simple():
    bucket_name, resp, trail_name = create_trail_simple()
    assert resp["Name"] == trail_name
    assert resp["S3BucketName"] == bucket_name
    assert "S3KeyPrefix" not in resp
    assert "SnsTopicName" not in resp
    assert "SnsTopicARN" not in resp
    assert resp["IncludeGlobalServiceEvents"] is True
    assert resp["IsMultiRegionTrail"] is False
    assert (
        resp["TrailARN"]
        == f"arn:aws:cloudtrail:us-east-1:{ACCOUNT_ID}:trail/{trail_name}"
    )
    assert resp["LogFileValidationEnabled"] is False
    assert resp["IsOrganizationTrail"] is False


def create_trail_simple(region_name="us-east-1"):
    client = boto3.client("cloudtrail", region_name=region_name)
    s3 = boto3.client("s3", region_name="us-east-1")
    bucket_name = str(uuid4())
    s3.create_bucket(Bucket=bucket_name)
    trail_name = str(uuid4())
    resp = client.create_trail(Name=trail_name, S3BucketName=bucket_name)
    return bucket_name, resp, trail_name


@mock_cloudtrail
def test_create_trail_multi_but_not_global():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.create_trail(
            Name="mytrailname",
            S3BucketName="non-existent",
            IncludeGlobalServiceEvents=False,
            IsMultiRegionTrail=True,
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidParameterCombinationException"
    # Note that this validation occurs before the S3 bucket is validated
    assert err["Message"] == "Multi-Region trail must include global service events."


@mock_cloudtrail
@mock_s3
@mock_sns
def test_create_trail_with_nonexisting_topic():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    s3 = boto3.client("s3", region_name="us-east-1")
    bucket_name = str(uuid4())
    s3.create_bucket(Bucket=bucket_name)
    with pytest.raises(ClientError) as exc:
        client.create_trail(
            Name="mytrailname",
            S3BucketName=bucket_name,
            SnsTopicName="nonexistingtopic",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InsufficientSnsTopicPolicyException"
    assert (
        err["Message"] == "SNS Topic does not exist or the topic policy is incorrect!"
    )


@mock_cloudtrail
@mock_s3
@mock_sns
def test_create_trail_advanced():
    bucket_name, resp, sns_topic_name, trail_name = create_trail_advanced()
    assert resp["Name"] == trail_name
    assert resp["S3BucketName"] == bucket_name
    assert resp["S3KeyPrefix"] == "s3kp"
    assert resp["SnsTopicName"] == sns_topic_name
    assert resp["SnsTopicARN"] == f"arn:aws:sns:us-east-1:{ACCOUNT_ID}:{sns_topic_name}"
    assert resp["IncludeGlobalServiceEvents"] is True
    assert resp["IsMultiRegionTrail"] is True
    assert (
        resp["TrailARN"]
        == f"arn:aws:cloudtrail:us-east-1:{ACCOUNT_ID}:trail/{trail_name}"
    )
    assert resp["LogFileValidationEnabled"] is True
    assert resp["IsOrganizationTrail"] is True
    assert resp["CloudWatchLogsLogGroupArn"] == "cwllga"
    assert resp["CloudWatchLogsRoleArn"] == "cwlra"
    assert resp["KmsKeyId"] == "kki"


def create_trail_advanced(region_name="us-east-1"):
    client = boto3.client("cloudtrail", region_name=region_name)
    s3 = boto3.client("s3", region_name="us-east-1")
    sns = boto3.client("sns", region_name=region_name)
    bucket_name = str(uuid4())
    s3.create_bucket(Bucket=bucket_name)
    sns_topic_name = "cloudtrailtopic"
    sns.create_topic(Name=sns_topic_name)
    trail_name = str(uuid4())
    resp = client.create_trail(
        Name=trail_name,
        S3BucketName=bucket_name,
        S3KeyPrefix="s3kp",
        SnsTopicName=sns_topic_name,
        IncludeGlobalServiceEvents=True,
        IsMultiRegionTrail=True,
        EnableLogFileValidation=True,
        IsOrganizationTrail=True,
        CloudWatchLogsLogGroupArn="cwllga",
        CloudWatchLogsRoleArn="cwlra",
        KmsKeyId="kki",
        TagsList=[{"Key": "tk", "Value": "tv"}, {"Key": "tk2", "Value": "tv2"}],
    )
    return bucket_name, resp, sns_topic_name, trail_name


@mock_cloudtrail
def test_get_trail_with_one_char():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.get_trail(Name="?")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidTrailNameException"
    assert (
        err["Message"]
        == "Trail name too short. Minimum allowed length: 3 characters. Specified name length: 1 characters."
    )


@mock_cloudtrail
def test_get_trail_unknown():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.get_trail(Name="unknowntrail")
    err = exc.value.response["Error"]
    assert err["Code"] == "TrailNotFoundException"
    assert err["Message"] == f"Unknown trail: unknowntrail for the user: {ACCOUNT_ID}"


@mock_cloudtrail
@mock_s3
def test_get_trail():
    create_trail_simple()
    client = boto3.client("cloudtrail", region_name="us-east-1")
    _, _, name = create_trail_simple()
    trail = client.get_trail(Name=name)["Trail"]
    assert trail["Name"] == name
    assert trail["IncludeGlobalServiceEvents"] is True
    assert trail["IsMultiRegionTrail"] is False
    assert (
        trail["TrailARN"] == f"arn:aws:cloudtrail:us-east-1:{ACCOUNT_ID}:trail/{name}"
    )


@mock_cloudtrail
def test_get_trail_status_with_one_char():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.get_trail_status(Name="?")
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidTrailNameException"
    assert (
        err["Message"]
        == "Trail name too short. Minimum allowed length: 3 characters. Specified name length: 1 characters."
    )


@mock_cloudtrail
def test_get_trail_status_unknown_trail():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.get_trail_status(Name="unknowntrail")
    err = exc.value.response["Error"]
    assert err["Code"] == "TrailNotFoundException"
    assert (
        err["Message"]
        == f"Unknown trail: arn:aws:cloudtrail:us-east-1:{ACCOUNT_ID}:trail/unknowntrail for the user: {ACCOUNT_ID}"
    )


@mock_cloudtrail
@mock_s3
def test_get_trail_status_inactive():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    _, _, trail_name = create_trail_simple()
    status = client.get_trail_status(Name=trail_name)
    assert status["IsLogging"] is False
    assert status["LatestDeliveryAttemptTime"] == ""
    assert status["LatestNotificationAttemptTime"] == ""
    assert status["LatestNotificationAttemptSucceeded"] == ""
    assert status["LatestDeliveryAttemptSucceeded"] == ""
    assert status["TimeLoggingStarted"] == ""
    assert status["TimeLoggingStopped"] == ""
    assert "StartLoggingTime" not in status


@mock_cloudtrail
@mock_s3
def test_get_trail_status_arn_inactive():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    _, resp, _ = create_trail_simple()
    status = client.get_trail_status(Name=resp["TrailARN"])
    assert status["IsLogging"] is False
    assert status["LatestDeliveryAttemptTime"] == ""
    assert status["LatestNotificationAttemptTime"] == ""
    assert status["LatestNotificationAttemptSucceeded"] == ""
    assert status["LatestDeliveryAttemptSucceeded"] == ""
    assert status["TimeLoggingStarted"] == ""
    assert status["TimeLoggingStopped"] == ""
    assert "StartLoggingTime" not in status


@mock_cloudtrail
@mock_s3
def test_get_trail_status_after_starting():
    client = boto3.client("cloudtrail", region_name="eu-west-3")
    _, _, trail_name = create_trail_simple(region_name="eu-west-3")

    client.start_logging(Name=trail_name)

    status = client.get_trail_status(Name=trail_name)
    assert status["IsLogging"] is True
    assert isinstance(status["LatestDeliveryTime"], datetime)
    assert isinstance(status["StartLoggingTime"], datetime)
    # .equal("2021-10-13T15:36:53Z")
    assert "LatestDeliveryAttemptTime" in status
    assert status["LatestNotificationAttemptTime"] == ""
    assert status["LatestNotificationAttemptSucceeded"] == ""
    # .equal("2021-10-13T15:36:53Z")
    assert "LatestDeliveryAttemptSucceeded" in status
    assert "TimeLoggingStarted" in status  # .equal("2021-10-13T15:02:21Z")
    assert status["TimeLoggingStopped"] == ""
    assert "StopLoggingTime" not in status


@mock_cloudtrail
@mock_s3
def test_get_trail_status_after_starting_and_stopping():
    client = boto3.client("cloudtrail", region_name="eu-west-3")
    _, _, trail_name = create_trail_simple(region_name="eu-west-3")

    client.start_logging(Name=trail_name)

    client.stop_logging(Name=trail_name)

    status = client.get_trail_status(Name=trail_name)
    assert status["IsLogging"] is False
    assert isinstance(status["LatestDeliveryTime"], datetime)
    assert isinstance(status["StartLoggingTime"], datetime)
    assert isinstance(status["StopLoggingTime"], datetime)
    # .equal("2021-10-13T15:36:53Z")
    assert "LatestDeliveryAttemptTime" in status
    assert status["LatestNotificationAttemptTime"] == ""
    assert status["LatestNotificationAttemptSucceeded"] == ""
    # .equal("2021-10-13T15:36:53Z")
    assert "LatestDeliveryAttemptSucceeded" in status
    assert "TimeLoggingStarted" in status  # .equal("2021-10-13T15:02:21Z")
    assert "TimeLoggingStopped" in status  # .equal("2021-10-13T15:03:21Z")


@mock_cloudtrail
@mock_s3
@mock_sns
def test_list_trails_different_home_region_one_multiregion():
    client = boto3.client("cloudtrail", region_name="eu-west-3")

    create_trail_simple()
    _, trail2, _, _ = create_trail_advanced(region_name="ap-southeast-2")
    create_trail_simple(region_name="eu-west-1")

    all_trails = client.list_trails()["Trails"]

    # Only the Trail created in the ap-southeast-2 is MultiRegion
    assert all_trails == [
        {
            "TrailARN": trail2["TrailARN"],
            "Name": trail2["Name"],
            "HomeRegion": "ap-southeast-2",
        }
    ]


@mock_cloudtrail
@mock_s3
@mock_sns
def test_list_trails_different_home_region_no_multiregion():
    client = boto3.client("cloudtrail", region_name="eu-west-3")

    create_trail_simple()
    create_trail_simple(region_name="ap-southeast-2")
    create_trail_simple(region_name="eu-west-1")

    all_trails = client.list_trails()["Trails"]
    # Since there is no MultiRegion Trail created
    # the eu-west-3 has no Trails
    assert len(all_trails) == 0


@mock_cloudtrail
@mock_s3
@mock_sns
def test_describe_trails_without_shadowtrails():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    _, trail1, _ = create_trail_simple()
    _, trail2, _, _ = create_trail_advanced()
    create_trail_simple(region_name="eu-west-1")

    # There are two Trails created in the us-east-1
    # one MultiRegion and the other is not MultiRegion
    trails = client.describe_trails()["trailList"]
    assert len(trails) == 2

    first_trail = [t for t in trails if t["Name"] == trail1["Name"]][0]
    assert first_trail["Name"] == trail1["Name"]
    assert first_trail["S3BucketName"] == trail1["S3BucketName"]
    assert first_trail["IncludeGlobalServiceEvents"] is True
    assert first_trail["IsMultiRegionTrail"] is False
    assert first_trail["HomeRegion"] == "us-east-1"
    assert first_trail["LogFileValidationEnabled"] is False
    assert first_trail["HasCustomEventSelectors"] is False
    assert first_trail["HasInsightSelectors"] is False
    assert first_trail["IsOrganizationTrail"] is False
    assert "S3KeyPrefix" not in first_trail
    assert "SnsTopicName" not in first_trail
    assert "SnsTopicARN" not in first_trail

    second_trail = [t for t in trails if t["Name"] == trail2["Name"]][0]
    assert second_trail["Name"] == trail2["Name"]
    assert second_trail["S3BucketName"] == trail2["S3BucketName"]
    assert second_trail["S3KeyPrefix"] == trail2["S3KeyPrefix"]
    assert second_trail["SnsTopicName"] == trail2["SnsTopicName"]
    assert second_trail["SnsTopicARN"] == trail2["SnsTopicARN"]
    assert second_trail["IncludeGlobalServiceEvents"] is True
    assert second_trail["IsMultiRegionTrail"] is True
    assert second_trail["HomeRegion"] == "us-east-1"
    assert second_trail["LogFileValidationEnabled"] is True
    assert second_trail["HasCustomEventSelectors"] is False
    assert second_trail["HasInsightSelectors"] is False
    assert second_trail["IsOrganizationTrail"] is True


@mock_cloudtrail
@mock_s3
@mock_sns
def test_describe_trails_with_shadowtrails_true():
    # Same behaviour as if shadowtrails-parameter was not supplied
    client = boto3.client("cloudtrail", region_name="us-east-1")
    create_trail_simple()
    create_trail_advanced()
    create_trail_simple(region_name="eu-west-1")

    # There are two Trails created in the us-east-1
    # one MultiRegion and the other is not MultiRegion
    trails = client.describe_trails(includeShadowTrails=True)["trailList"]
    assert len(trails) == 2

    # There are two Trails in the eu-west-1
    # one MultiRegion (created in the us-east-1)
    # and another not MultiRegion created in the us-east-1
    eu_client = boto3.client("cloudtrail", region_name="eu-west-1")
    trails = eu_client.describe_trails(includeShadowTrails=True)["trailList"]
    assert len(trails) == 2


@mock_cloudtrail
@mock_s3
@mock_sns
def test_describe_trails_with_shadowtrails_false():
    # Only trails for the current region should now be returned
    client = boto3.client("cloudtrail", region_name="us-east-1")
    _, _, name1 = create_trail_simple()
    _, _, _, name2 = create_trail_advanced()
    _, _, name3 = create_trail_simple(region_name="eu-west-1")

    trails = client.describe_trails(includeShadowTrails=False)["trailList"]
    assert len(trails) == 2
    assert [t["Name"] for t in trails] == [name1, name2]

    eu_client = boto3.client("cloudtrail", region_name="eu-west-1")
    trails = eu_client.describe_trails(includeShadowTrails=False)["trailList"]
    assert len(trails) == 1
    assert [t["Name"] for t in trails] == [name3]


@mock_cloudtrail
@mock_s3
def test_delete_trail():
    client = boto3.client("cloudtrail", region_name="us-east-1")
    _, _, name = create_trail_simple()

    trails = client.describe_trails()["trailList"]
    assert len(trails) == 1

    client.delete_trail(Name=name)

    trails = client.describe_trails()["trailList"]
    assert len(trails) == 0


@mock_cloudtrail
@mock_s3
def test_update_trail_simple():
    client = boto3.client("cloudtrail", region_name="ap-southeast-2")
    bucket_name, trail, name = create_trail_simple(region_name="ap-southeast-2")
    resp = client.update_trail(Name=name)
    assert resp["Name"] == name
    assert resp["S3BucketName"] == bucket_name
    assert resp["IncludeGlobalServiceEvents"] is True
    assert resp["IsMultiRegionTrail"] is False
    assert resp["LogFileValidationEnabled"] is False
    assert resp["IsOrganizationTrail"] is False
    assert "S3KeyPrefix" not in resp
    assert "SnsTopicName" not in resp
    assert "SnsTopicARN" not in resp

    trail = client.get_trail(Name=name)["Trail"]
    assert trail["Name"] == name
    assert trail["S3BucketName"] == bucket_name
    assert trail["IncludeGlobalServiceEvents"] is True
    assert trail["IsMultiRegionTrail"] is False
    assert trail["LogFileValidationEnabled"] is False
    assert trail["IsOrganizationTrail"] is False
    assert "S3KeyPrefix" not in trail
    assert "SnsTopicName" not in trail
    assert "SnsTopicARN" not in trail


@mock_cloudtrail
@mock_s3
def test_update_trail_full():
    client = boto3.client("cloudtrail", region_name="ap-southeast-1")
    _, trail, name = create_trail_simple(region_name="ap-southeast-1")
    resp = client.update_trail(
        Name=name,
        S3BucketName="updated_bucket",
        S3KeyPrefix="s3kp",
        SnsTopicName="stn",
        IncludeGlobalServiceEvents=False,
        IsMultiRegionTrail=True,
        EnableLogFileValidation=True,
        CloudWatchLogsLogGroupArn="cwllga",
        CloudWatchLogsRoleArn="cwlra",
        KmsKeyId="kki",
        IsOrganizationTrail=True,
    )
    assert resp["Name"] == name
    assert resp["S3BucketName"] == "updated_bucket"
    assert resp["S3KeyPrefix"] == "s3kp"
    assert resp["SnsTopicName"] == "stn"
    assert resp["IncludeGlobalServiceEvents"] is False
    assert resp["IsMultiRegionTrail"] is True
    assert resp["LogFileValidationEnabled"] is True
    assert resp["IsOrganizationTrail"] is True

    trail = client.get_trail(Name=name)["Trail"]
    assert trail["Name"] == name
    assert trail["S3BucketName"] == "updated_bucket"
    assert trail["S3KeyPrefix"] == "s3kp"
    assert trail["SnsTopicName"] == "stn"
    assert trail["IncludeGlobalServiceEvents"] is False
    assert trail["IsMultiRegionTrail"] is True
    assert trail["LogFileValidationEnabled"] is True
    assert trail["IsOrganizationTrail"] is True
    assert trail["CloudWatchLogsLogGroupArn"] == "cwllga"
    assert trail["CloudWatchLogsRoleArn"] == "cwlra"
    assert trail["KmsKeyId"] == "kki"
